#  Copyright (C) 2021, Christoph Peters, Karlsruhe Institute of Technology
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.


import numpy as np
from matplotlib import pyplot


def load_doubles(file, count):
    """
    Loads the given number of doubles from the given File. If count is greater
    than one, the result is returned as array, otherwise as scalar.
    """
    doubles = np.fromfile(file, np.longdouble, count)
    doubles = [float(double) for double in doubles]
    return np.asarray(doubles) if count > 1 else doubles[0]


class TestCase:
    """
    A test case for the method that samples a sector in the area between
    ellipses.
    """

    def __init__(self, file_path):
        """
        Loads the given test case produced by the find_worst_case program.
        """
        with open(file_path, "rb") as file:
            # The parameters from which the test case has been constructed
            self.parameters = load_doubles(file, 6)
            # The inner and outer ellipse
            self.inner_ellipse = load_doubles(file, 2)
            self.outer_ellipse = load_doubles(file, 2)
            # The direction vectors bounding the sector
            self.dir_0 = load_doubles(file, 2)
            self.dir_1 = load_doubles(file, 2)
            # The random number number used to sample the azimuth
            self.random_number = load_doubles(file, 1)
            # The area between the ellipses within the sector
            self.sector_area = load_doubles(file, 1)
            # The sampled direction
            self.sample_dirs = [load_doubles(file, 2) for i in range(3)]

    def plot(self):
        """
        Creates a new figure visualizing this test case. Returns the figure.
        """
        figure = pyplot.figure()
        pyplot.subplot(aspect=1.0)
        ellipses = [self.inner_ellipse, self.outer_ellipse]
        phase = np.linspace(-np.pi, np.pi, 10001)
        circle = np.vstack([np.cos(phase), np.sin(phase)]).T
        pyplot.plot(circle[:, 0], circle[:, 1], color=[0.9] * 3)
        for i, ellipse in enumerate(ellipses):
            factor = 1.0 / (np.sqrt(ellipse @ ellipse + 1.0) + 1.0)
            inv_sqrt = np.linalg.inv(np.eye(2) + factor * np.outer(ellipse, ellipse))
            ellipse_samples = (inv_sqrt @ circle.T).T
            pyplot.plot(ellipse_samples[:, 0], ellipse_samples[:, 1], ":", color="C%d" % i, linewidth=3.0)
        for i, dir in enumerate([self.dir_0, self.dir_1] + self.sample_dirs):
            dir = dir / np.linalg.norm(dir)
            dir *= 1.0 + 0.1 * i
            pyplot.plot([0.0, dir[0]], [0.0, dir[1]], color="C%d" % i, zorder=-i)
        return figure

    def print(self):
        """Prints information about this test case."""
        print("Parameters:", self.parameters)
        print("Inner ellipse:", self.inner_ellipse)
        print("Outer ellipse:", self.outer_ellipse)
        print("Sector dirs:", self.dir_0, self.dir_1)
        print("Sector area:", self.sector_area)
        print("Random number:", self.random_number)


if __name__ == "__main__":
    test = TestCase("find_worst_case/test_case.dat")
    test.plot()
    test.print()
    pyplot.show()
