#  Copyright (C) 2021, Christoph Peters, Karlsruhe Institute of Technology
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.


import numpy as np
from polygon_clipping import clip_polygon


def clip_polygon_reference(polygon):
    """
    Clips the given polygon against the horizon.
    :param polygon: A (vertex_count, 3) array providing a polygon.
    :return: A (clipped_vertex_count, 3) array providing the clipped polygon.
    """
    vertex_count = polygon.shape[0]
    # Clip the polygon against the horizon
    polygon_z = polygon[:, 2]
    edges = np.roll(polygon, 1, 0) - polygon
    edges_z = edges[:, 2]
    edge_intersections = -polygon_z / edges[:, 2]
    clipped_polygon = np.zeros((vertex_count + 1, 3))
    clipped_vertex_count = 0
    for j in range(vertex_count):
        if polygon_z[j] > 0.0 and edges_z[j] > 0.0:
            clipped_polygon[clipped_vertex_count, :] = polygon[j, :]
            clipped_vertex_count += 1
        if 0.0 < edge_intersections[j] < 1.0:
            clipped_polygon[clipped_vertex_count, :] = polygon[j, :] + edge_intersections[j] * edges[j, :]
            clipped_vertex_count += 1
        if polygon_z[j] > 0.0 and edges_z[j] <= 0.0:
            clipped_polygon[clipped_vertex_count, :] = polygon[j, :]
            clipped_vertex_count += 1
    return clipped_polygon[:clipped_vertex_count, :]


def generate_test_case(vertex_count, count_above, first_above):
    """
    Returns a polygon as (vertex_count, 3) array where count_above vertices
    have positive z-coordinate and the lowest index with positive z-coordinate
    is first_above.
    """
    # Generate count_above vertices above the horizon
    polygon = [(0.0, np.cos(np.pi * i / count_above), 1.0 + np.sin(np.pi * i / count_above)) for i in range(count_above)]
    # Generate the remaining vertices below the horizon
    count_below = vertex_count - count_above
    polygon += [(0.0, np.cos(np.pi + np.pi * i / count_below), -1.0 + np.sin(np.pi + np.pi * i / count_below)) for i in range(count_below)]
    # Shear a bit
    polygon = np.asarray(polygon)
    polygon[:, 0] += 2.3 * polygon[:, 1]
    # Add a bit of noise
    polygon += 1.0e-2 * np.random.rand(vertex_count, 3)
    # Now apply a cyclic shift
    return np.roll(polygon, first_above, 0)


def run_test(vertex_count, count_above, first_above, array_tight):
    """
    Generates a test case using generate_test_case() with the given parameters
    to see if clip_polygon_reference() and clip_polygon() produce consistent
    results. Prints information about issues.
    """
    polygon = generate_test_case(vertex_count, count_above, first_above)
    clipped_polygon_reference = clip_polygon_reference(np.copy(polygon))
    max_vertex_count = vertex_count + (1 if array_tight else 2)
    clipped_vertex_count, clipped_polygon = clip_polygon(np.copy(polygon), max_vertex_count)
    # Verify that the first vertex is repeated properly and remove
    # it
    if clipped_vertex_count < max_vertex_count:
        if not np.all(clipped_polygon[0, :] == clipped_polygon[clipped_vertex_count, :]):
            print("Last vertex mismatch.")
            print("The test case is: %d, %d, %d, " % (vertex_count, count_above, first_above) + str(array_tight))
    clipped_polygon = clipped_polygon[:clipped_vertex_count]
    # Compare vertex counts
    if clipped_polygon.shape[0] != clipped_polygon_reference.shape[0]:
        print("Vertex count does not match. GLSL says %d, reference says %d." % (
        clipped_polygon.shape[0], clipped_polygon_reference.shape[0]))
        print("The test case is: %d, %d, %d, " % (vertex_count, count_above, first_above) + str(array_tight))
    # One cyclic shift must match
    matches = np.any([np.linalg.norm(np.roll(clipped_polygon_reference, offset, 0) - clipped_polygon) < 1.0e-12
                      for offset in range(clipped_polygon.shape[0])])
    if not matches and clipped_polygon.shape[0] > 0:
        print("Clipped polygon mismatch.")
        print("The test case is: %d, %d, %d, " % (vertex_count, count_above, first_above) + str(array_tight))


if __name__ == "__main__":
    # Go through all possible cases
    for vertex_count in range(3, 8):
        for count_above in range(vertex_count + 1):
            for first_above in range(vertex_count):
                for array_tight in [False, True]:
                    run_test(vertex_count, count_above, first_above, array_tight)
    print("All tests ran. If there were problems, they have been listed above.")
