#  Copyright (C) 2021, Christoph Peters, Karlsruhe Institute of Technology
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.


import numpy as np
from scipy import interpolate
import re
from matplotlib import pyplot
from os import path, walk
import imageio
from subprocess import call


class IESProfile:
    """
    This class represents photometric data from an IES profile. It is able to
    load it from a file and can store it to a texture.
    """

    def __init__(self, file_path):
        """
        Loads an IES profile from the *.ies file at the given path. Upon
        success the member valid is set to True. Follows the description from:
        https://docs.agi32.com/AGi32/Content/references/IDH_iesna_standard_file_format.htm
        """
        self.valid = False
        text = open(file_path, "r").read()
        # Figure out what type of IES file we are dealing with
        self.type = None
        for type in ["IESNA91", "IESNA:LM-63-1995", "IESNA:LM-63-2002"]:
            if re.match("^" + type, text) is not None:
                self.type = type
        if self.type is None:
            print("The file at path %s does not appear to be a valid IES profile." % file_path)
            return
        if self.type == "IESNA91":
            print("The file at path %s uses standard IESNA91, which is not supported." % file_path)
            return
        # Load attributes
        for attribute_match in re.finditer(r"\[([_a-zA-Z]\w+)\]\s+(.*)", text):
            self.__dict__[attribute_match.group(1)] = attribute_match.group(2)
        # Find the tilt keyword
        tilt_match = re.search(r"TILT\s*=\s*(\w+)\s*", text)
        if tilt_match is None:
            print("The file at path %s does not appear to be a valid IES profile." % file_path)
            return
        if tilt_match.group(1) != "NONE":
            print(
                "The file at path %s uses an unsupported tilt (should be NONE): %s" % (file_path, tilt_match.group(1)))
            return
        # Now we only care about the part after tilt
        offset = tilt_match.end()
        # Read meta data. We specify the format by simply giving names to all
        # integer (i_) or floating point (f_) attributes
        attributes = [
            "i_lamp_count", "f_lumens", "f_multiplier", "i_vertical_angle_count", "i_horizontal_angle_count",
            "i_photometric_type", "i_units_type", "f_width", "f_length", "f_height",
            "f_ballast_factor", "f_unused", "f_input_watts"
        ]
        int_re = r"[+-]?[0-9]+"
        float_re = r"[+-]?([0-9]*\.)?[0-9]+"
        for attribute in attributes:
            name = attribute[2:]
            is_int = (attribute[0:2] == "i_")
            match = re.search(int_re if is_int else float_re, text[offset:])
            if match is None:
                print("Failed to locate the attribute %s in the following text." % attribute)
                print(text[offset:])
            value = int(match.group(0)) if is_int else float(match.group(0))
            offset += match.end()
            if name == "vertical_angle_count":
                vertical_angle_count = value
            elif name == "horizontal_angle_count":
                horizontal_angle_count = value
            else:
                self.__dict__[name] = value
        # What remains of the file is just floats
        floats = np.asarray([float(match.group(0)) for match in re.finditer(float_re, text[offset:])])
        expected_float_count = vertical_angle_count + horizontal_angle_count
        expected_float_count += horizontal_angle_count * vertical_angle_count
        if floats.size < expected_float_count:
            print("The file at path %s should carry %d floating point values after the meta-data but carries %d."
                  % (file_path, floats.size, expected_float_count))
            return
        self.vertical_angles = floats[0:vertical_angle_count]
        float_offset = vertical_angle_count
        self.horizontal_angles = floats[float_offset:float_offset + horizontal_angle_count]
        float_offset += horizontal_angle_count
        self.intensities = floats[float_offset:].reshape((horizontal_angle_count, vertical_angle_count)).T
        # Apply the multiplier
        self.intensities *= self.multiplier
        self.valid = True

    def make_radially_symmetric(self):
        """
        Takes the mean across horizontal angles to make this profile radially
        symmetric with respect to the vertical axis.
        """
        self.horizontal_angles = np.asarray([0.0])
        self.intensities = np.mean(self.intensities, axis=1)[:, np.newaxis]

    def complete(self):
        """
        There are conventions for how to represent symmetries or occlusion in
        emission profiles. If vertical angles cover only 90 degrees, one
        hemisphere receives no light. If horizontal angles cover less than 360
        degrees, data is repeated. This function implements these conventions
        changing the arrays such that they cover vertical angles from 0 to 180
        degrees and horizontal angles from 0 to 360 degrees. It does not
        guarantee an equidistant sampling for the output.
        """
        # Set the absent hemisphere to zero
        if self.vertical_angles[0] == 0.0 and self.vertical_angles[-1] == 90.0:
            self.vertical_angles = np.concatenate([self.vertical_angles, [90.0 + 1.0e-6, 180.0]])
            self.intensities = np.vstack([self.intensities, np.zeros((2, self.intensities.shape[1]))])
        if self.vertical_angles[0] == 90.0 and self.vertical_angles[-1] == 180.0:
            self.vertical_angles = np.concatenate([[0.0, 90.0 - 1.0e-6], self.vertical_angles])
            self.intensities = np.vstack([np.zeros((2, self.intensities.shape[1])), self.intensities])
        # Apply repetition for horizontal angles
        stride = self.horizontal_angles[-1]
        if stride == 90.0:
            repetitions = 4
        elif stride == 180.0:
            repetitions = 2
        else:
            repetitions = 1
        self.horizontal_angles = np.concatenate([self.horizontal_angles] + [self.horizontal_angles[1:] + j * stride for j in range(1, repetitions)])
        self.intensities = np.hstack([self.intensities] + [self.intensities[:, 1:] for j in range(1, repetitions)])
        # Radially symmetric profiles get 2 samples to ease resampling
        if self.horizontal_angles.size == 1:
            self.horizontal_angles = np.asarray([0.0, 360.0])
            self.intensities = np.tile(self.intensities, (1, 2))

    def plot(self, plot_grid=True):
        """
        Uses matplotlib to plot this IES profile to the current axes. If the
        profile is 1D, it is plotted as angular graph, otherwise it is plotted
        as image. For 1D plots, you can optionally request an angular grid.
        """
        if self.horizontal_angles.size == 1:
            if plot_grid:
                grid_color = 3 * [0.7]
                phase = np.linspace(-np.pi, np.pi, 1001)
                for radius in np.linspace(0.0, np.max(self.intensities), 4)[1:]:
                    pyplot.plot(radius * np.cos(phase), radius * np.sin(phase), color=grid_color)
                for sector_angles in np.arange(0.0, np.pi, 0.25 * np.pi):
                    pyplot.plot([-radius * np.cos(sector_angles), radius * np.cos(sector_angles)],
                                [-radius * np.sin(sector_angles), radius * np.sin(sector_angles)],
                                color=grid_color)
            vertical_radians = self.vertical_angles * np.pi / 180.0
            for sign in [-1.0, 1.0]:
                pyplot.plot(sign * np.sin(vertical_radians) * self.intensities[:, 0],
                            -np.cos(vertical_radians) * self.intensities[:, 0], "C0")
        else:
            pyplot.imshow(self.intensities, vmin=0.0)
            # Define an appropriate number of ticks
            for axis, angles in [(pyplot.gca().get_xaxis(), self.horizontal_angles), (pyplot.gca().get_yaxis(), self.vertical_angles)]:
                stride = 1
                while angles.size // stride > 12:
                    stride *= 2
                axis.set_ticks(np.arange(0.0, angles.size, stride))
                axis.set_ticklabels(["%.0f" % angle for angle in angles[::stride]])
            pyplot.xlabel("Horizontal angles")
            pyplot.ylabel("Vertical angles")
            pyplot.colorbar()

    def create_texture(self, normalize=True, resolution_scale=1.0):
        """
        Creates a from this IES profile. The texture is guaranteed to have a
        power of two resolution and to sample the range from 0 to 360
        (horizontal) or 180 (vertical) degrees equidistantly. You should invoke
        complete() before invoking this method.
        :param normalize: Pass True to ensure that the integral of the returned
            emission profile over the unit sphere is 4.0 * np.pi.
        :param resolution_scale: The resolution is multiplied by this factor
            before rounding up to a power of two.
        :return: A 2D array of shape (vertical_angle_count,
            horizontal_angle_count) holding floating point luminance values.
            Samples are placed in a manner that is consistent with bilinear
            texture interpolation in GLSL, i.e. turning angles into texture
            coordinates works fine.
        """
        if self.horizontal_angles[-1] != 360.0 or self.vertical_angles[-1] != 180.0:
            print("You have to invoke complete() before you can create a texture from an IES profile.")
            return None
        # Decide what resolution to use (smallest power of two greater than
        # what we have)
        extent = np.asarray(2.0 ** np.ceil(np.log2(np.asarray(self.intensities.shape, dtype=np.float) * resolution_scale)), dtype=np.int)
        # Prepare for resampling with bilinear interpolation
        interpolator = interpolate.RectBivariateSpline(self.horizontal_angles, self.vertical_angles, self.intensities.T, kx=1, ky=1)
        # Sample at the appropriate locations
        horizontal_angles = np.arange(0.5 * 360.0 / extent[1], 360.0, 360.0 / extent[1])
        vertical_angles = np.arange(0.5 * 180.0 / extent[0], 180.0, 180.0 / extent[0])
        result = interpolator(horizontal_angles, vertical_angles).T
        # Apply normalization
        if normalize:
            sin_theta = np.sin(vertical_angles * np.pi / 180.0)[:, np.newaxis].repeat(horizontal_angles.size, 1)
            result *= np.sum(sin_theta) / np.sum(result * sin_theta)
        return result

    def store_texture(self, output_path, *create_texture_args):
        """
        Invokes complete(), forwards to create_texture() and writes the result
        to the given file using imageio.
        """
        self.complete()
        texture = self.create_texture(*create_texture_args)
        rgb_image = np.asarray(texture[:, :, np.newaxis], dtype=np.float32).repeat(3, 2)
        # If this line fails, try running the folling call once:
        # imageio.plugins.freeimage.download()
        imageio.imwrite(output_path, rgb_image)


if __name__ == "__main__":
    for directory, directories, files in walk("../Models/ies_profiles"):
        for file in files:
            if path.splitext(file)[1].lower() != ".ies":
                continue
            ies_file_path = path.join(directory, file)
            ies_profile = IESProfile(ies_file_path)
            if not ies_profile.valid:
                continue
            if False:
                pyplot.figure()
                pyplot.subplot(1, 1, 1, aspect=1.0, title=ies_file_path)
                ies_profile.plot()
                pyplot.show()
            # Convert to a HDR texture
            hdr_file_path = path.splitext(ies_file_path)[0] + ".hdr"
            ies_profile.store_texture(hdr_file_path)
            # Convert to a texture for use by the renderer
            texture_conversion_path = "../vulkan_renderer/tools/texture_conversion/build/texture_conversion"
            vkt_file_path = "../vulkan_renderer/data/ies_profiles/%s.vkt" % path.splitext(path.split(hdr_file_path)[1])[0]
            call([texture_conversion_path, "97", hdr_file_path, vkt_file_path])
            print("Created %s." % vkt_file_path)
    print("Done.")
