//  Copyright (C) 2021, Christoph Peters, Karlsruhe Institute of Technology
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <https://www.gnu.org/licenses/>.


#include "nelder_mead.h"
#include "polygon_sampling_ported.h"
#include <iostream>
#include <algorithm>
#include <fstream>
#include <random>
#include <cstdint>
#include <thread>
#include <mutex>
#include <utility>
#include <chrono>
#include <memory>


//! Maps numbers between zero and one to numbers between -infty and infty in a
//! bijective and monotonic manner
inline Float map_to_all_reals(Float number_between_zero_and_one) {
	return tan((0.5 * M_PI) * (number_between_zero_and_one * 2.0 - 1.0));
}

//! Maps numbers between zero and one to numbers between 0 and infty in a
//! bijective and monotonic manner
inline Float map_to_positive_reals(Float number_between_zero_and_one) {
	return tan((0.5 * M_PI) * number_between_zero_and_one);
}

//! Describes a single test case for the sampling procedure, i.e. a
//! configuration in which one sample can be produced
struct test_case_t {
	//! The parameter values from which this test case has been constructed
	search_vec_t parameters;
	//! The inner and outer ellipse
	vec2 inner_ellipse, outer_ellipse;
	//! The direction vectors bounding the sector
	vec2 dir_0, dir_1;
	//! The random number number used to sample the azimuth
	Float random_number;
	//! The area between the ellipses within the sector
	Float sector_area;
	//! The sampled direction after 0, 1 and 2 iterations
	vec2 sample_dirs[3];
};


//! Produces a sample for the given test case
vec2 sample_test_case(test_case_t test, uint iteration_count=2) {
	Float target_area = test.random_number * test.sector_area;
	return sample_sector_between_ellipses(
		make_vec2(test.random_number, 1.0), target_area,
		test.inner_ellipse, test.outer_ellipse, test.dir_0, test.dir_1, iteration_count);
}


//! Turns a vector of six numbers between zero and one into a test case for the
//! sampling procedure
inline test_case_t construct_test_case(search_vec_t input, bool compute_sample_dirs=false) {
	test_case_t test;
	test.parameters = input;
	// Clamp the input
	for (uint i = 0; i != DIMENSIONALITY; ++i)
		input[i] = clamp(input[i], 0.0, 1.0);
	// Construct two ellipses. WLOG, the outer ellipse is symmetric with
	// respect to the x-axis (otherwise we can rotate everything).
	test.inner_ellipse = make_vec2(map_to_all_reals(input[0]), map_to_positive_reals(input[1]));
	test.outer_ellipse = make_vec2(0.0, map_to_positive_reals(input[2]));
	// Figure out where they intersect
	vec2 intersections[] = {
		normalize(solve_homogeneous_quadratic(outerProduct(test.inner_ellipse, test.inner_ellipse) - outerProduct(test.outer_ellipse, test.outer_ellipse))),
		normalize(solve_homogeneous_quadratic(outerProduct(test.outer_ellipse, test.outer_ellipse) - outerProduct(test.inner_ellipse, test.inner_ellipse)))
	};
	// We care about the two intersections above the x-axis
	for (uint i = 0; i != 2; ++i)
		intersections[i] = (intersections[i].y < 0.0) ? -intersections[i] : intersections[i];
	// Among those, we care about the one with the greater x-coordinate,
	// because we let the sector begin at the positive x-axis
	intersections[1] = (intersections[1].x < intersections[0].x) ? intersections[0] : intersections[1];
	intersections[0] = make_vec2(1.0, 0.0);
	// Place sector boundaries
	Float lerp_0 = input[3];
	Float lerp_1 = (1.0 - input[4]) * lerp_0 + input[4] * 1.0;
	Float intersection_angles[2];
	for (uint i = 0; i != 2; ++i)
		intersection_angles[i] = atan2(intersections[i].x, intersections[i].y);
	Float angle_0 = (1.0 - lerp_0) * intersection_angles[0] + lerp_0 * intersection_angles[1];
	Float angle_1 = (1.0 - lerp_1) * intersection_angles[0] + lerp_1 * intersection_angles[1];
	// The sector must be a bit smaller than 180 degrees
	const Float max_sector_angle_in_degrees = 179.999;
	const Float max_sector_angle_in_radians = max_sector_angle_in_degrees * M_PI / 180.0;
	if (angle_1 < angle_0 - max_sector_angle_in_radians)
		angle_1 = angle_0 - max_sector_angle_in_radians;
	test.dir_0 = make_vec2(sin(angle_0), cos(angle_0));
	test.dir_1 = make_vec2(sin(angle_1), cos(angle_1));
	// Decide where to take a sample
	test.random_number = input[5];
	test.sector_area = get_area_between_ellipses_in_sector(
		test.inner_ellipse, get_ellipse_rsqrt_det(test.inner_ellipse),
		test.outer_ellipse, get_ellipse_rsqrt_det(test.outer_ellipse),
		test.dir_0, test.dir_1);
	// In theory, our sampling procedure is invariant under mirroring of the
	// sector. In practice, there are slight cancellations for random_number
	// near 1.0 but not near 0.0. To circumvent them, we mirror the whole
	// configuration if random_number > 0.5
	if (test.random_number > 0.5) {
		test.inner_ellipse.x *= -1.0;
		test.outer_ellipse.x *= -1.0;
		test.dir_0.x *= -1.0;
		test.dir_1.x *= -1.0;
		swap(test.dir_0, test.dir_1);
		test.random_number = 1.0 - test.random_number;
	}
	// Perform sampling
	if (compute_sample_dirs)
		for (uint i = 0; i != 3; ++i)
			test.sample_dirs[i] = sample_test_case(test, i);
	return test;
}


//! Holds the worst error encountered by a thread and meta-data about the
//! effort invested to find it
struct thread_progress_t {
public:
	//! Initializes with zero error and no effort
	thread_progress_t() :
	worst_error(0.0),
	invocation_count(0),
	init_count(0) {
	}

	//! Adds up effort and takes the worst error between this object and the
	//! given one
	void merge(const thread_progress_t& rhs) {
		if (worst_error < rhs.worst_error) {
			worst_error = rhs.worst_error;
			worst_parameters = rhs.worst_parameters;
		}
		invocation_count += rhs.invocation_count;
		init_count += rhs.init_count;
	}

	//! The worst error encountered by this thread thus far
	Float worst_error;
	//! The parameters that lead to the worst error encountered thus far
	search_vec_t worst_parameters;
	//! The number of objective function invocations that this thread has
	//! performed
	uint64_t invocation_count;
	//! The number of random initializations that this thread has used
	uint64_t init_count;
};


//! A thread that will search worst case errors indefinitely
class thread_t : public thread {
public:
	//! Launches the thread
	thread_t(random_device& random_device) :
	random_generator(random_device()),
	uniform_distribution(0.0, 1.0) {
		// Launch the thread
		thread::operator=(thread(
			[this]() {
				while (true)
					try_random_initialization();
			}));
	}

	//! Generates and tries a single random initialization and updates the
	//! progress
	void try_random_initialization() {
		// Generate a random initialization
		search_vec_t initialization;
		for (uint i = 0; i != DIMENSIONALITY; ++i)
			initialization[i] = uniform_distribution(random_generator);
		// Use Nelder-Mead to maximize the error
		uint64_t invocation_count = 0;
		auto objective = [this, &invocation_count](const search_vec_t& point) -> Float {
			++invocation_count;
			test_case_t test = construct_test_case(point);
			// With tiny polygons, we run into numerical issues but this
			// experiment is concerned with theoretical convergence properties
			if (test.sector_area < 1.0e-2)
				return 0.0;
			// Take a sample
			vec2 sample = sample_test_case(test);
			// Compute the error of the result
			Float result_area = get_area_between_ellipses_in_sector(
				test.inner_ellipse, get_ellipse_rsqrt_det(test.inner_ellipse),
				test.outer_ellipse, get_ellipse_rsqrt_det(test.outer_ellipse),
				test.dir_0, sample);
			Float target_area = test.sector_area * test.random_number;
			Float result = abs((result_area - target_area) / test.sector_area);
			return -result;
		};
		Float error;
		search_vec_t optimized = minimize_nelder_mead(&error, initialization, objective);
		error = -error;
		// Update the progress
		progress_mutex.lock();
		progress.invocation_count += invocation_count;
		++progress.init_count;
		if (error > progress.worst_error) {
			progress.worst_error = error;
			progress.worst_parameters = optimized;
		}
		progress_mutex.unlock();
	}

	//! A random generator with a seed specific to this thread
	mt19937_64 random_generator;
	//! A uniform distribution on [0,1)
	uniform_real_distribution<Float> uniform_distribution;
	//! A mutex that should be acquired when reading or writing attributes of
	//! progress
	mutex progress_mutex;
	//! The progress of this thread
	thread_progress_t progress;
};


int main() {
	// Warn about Visual C++
	if (sizeof(long double) < 10) {
		cout << "WARNING: This program needs to be compiled in an environment where long double" << endl;
		cout << "uses at least 63 bits for the mantissa. Do not use Visual C++." << endl;
	}
	// Prepare truly random bits
	random_device random_device;

	// Take a sample with the worst case found previously to enable analysis
	if (false) {
		ifstream test_case_file("test_case.dat", ios_base::binary);
		if (test_case_file.is_open()) {
			test_case_t test;
			test_case_file.read(reinterpret_cast<char*>(&test), sizeof(test));
			test = construct_test_case(test.parameters);
			vec2 sample = sample_test_case(test);
			Float result_area = get_area_between_ellipses_in_sector(
				test.inner_ellipse, get_ellipse_rsqrt_det(test.inner_ellipse),
				test.outer_ellipse, get_ellipse_rsqrt_det(test.outer_ellipse),
				test.dir_0, sample);
			Float target_area = test.sector_area * test.random_number;
			Float error = abs((result_area - target_area) / test.sector_area);
			cout << "Error in a previous run: " << error << endl;
		}
	}
	// Compute the median error
	if (false) {
		// Prepare a random generator
		mt19937_64 random_generator(random_device());
		uniform_real_distribution<Float> uniform_distribution(0.0f, 1.0f);
		// Generate samples with random test cases and record their errors
		uint iteration_count = 2;
		size_t sample_count = 100000000;
		vector<Float> errors(sample_count);
		std::cout << "Using " << iteration_count << " iterations and " << sample_count << " samples to compute percentiles of errors." << std::endl;
		for (size_t i = 0; i != sample_count; ++i) {
			search_vec_t random_numbers;
			for (uint j = 0; j != DIMENSIONALITY; ++j)
				random_numbers[j] = uniform_distribution(random_generator);
			test_case_t test = construct_test_case(random_numbers);
			vec2 sample = sample_test_case(test, iteration_count);
			Float result_area = get_area_between_ellipses_in_sector(
				test.inner_ellipse, get_ellipse_rsqrt_det(test.inner_ellipse),
				test.outer_ellipse, get_ellipse_rsqrt_det(test.outer_ellipse),
				test.dir_0, sample);
			Float target_area = test.sector_area * test.random_number;
			errors[i] = abs((result_area - target_area) / test.sector_area);
		}
		// Sort and report percentiles
		uint percentiles_times_ten[] = {1, 2, 3, 4, 5, 10, 20, 30, 40, 50, 100, 200, 300, 400, 500, 600, 700, 800, 900, 950, 960, 970, 980, 990, 995, 996, 997, 998, 999};
		std::cout << "Sorting the errors." << std::endl;
		sort(errors.begin(), errors.end());
		for (uint percentile_times_ten : percentiles_times_ten) {
			size_t index = (percentile_times_ten * sample_count) / 1000;
			std::cout << percentile_times_ten * 0.1 << "th percentile of backward error: " << errors[index] << std::endl;
		}
	}

	// Run as many threads as the hardware can handle
	vector<unique_ptr<thread_t> > threads;
	uint thread_count = thread::hardware_concurrency();
	for (uint i = 0; i != thread_count; ++i)
		threads.emplace_back(new thread_t(random_device));
	// Report on their progress in regular intervals
	Float worst_error = 0.0;
	uint64_t reported_invocation_count = 0;
	while (true) {
		// Combine the progress of all threads
		thread_progress_t total_progress;
		for (const unique_ptr<thread_t>& thread : threads) {
			thread->progress_mutex.lock();
			total_progress.merge(thread->progress);
			thread->progress_mutex.unlock();
		}
		// Report the error if it is worse than anything seen thus far
		if (total_progress.worst_error > worst_error) {
			worst_error = total_progress.worst_error;
			cout << "Error: " << worst_error << endl;
			cout << "Test case:";
			for (uint i = 0; i != DIMENSIONALITY; ++i)
				cout << " " << total_progress.worst_parameters[i];
			cout << endl;
			cout << total_progress.invocation_count / 1000000 << " million invocations and " << total_progress.init_count << " initializations." << endl;
			test_case_t test = construct_test_case(total_progress.worst_parameters, true);
			ofstream("test_case.dat", ios_base::binary).write(reinterpret_cast<char*>(&test), sizeof(test));
		}
		// Report the invocation count and the initialization count regularly
		if (total_progress.invocation_count / 5000000000 > reported_invocation_count) {
			cout << total_progress.invocation_count / 1000000000 << " billion invocations and " << total_progress.init_count << " initializations." << endl;
			++reported_invocation_count;
		}
		// This thread should not stress the CPU
		this_thread::sleep_for(chrono::seconds());
	}
}
