//  Copyright (C) 2021, Christoph Peters, Karlsruhe Institute of Technology
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <https://www.gnu.org/licenses/>.


#include "glsl_math.h"

//! The dimensionality of the search space
#define DIMENSIONALITY 6
//! The vector type used to represent points in the search space
typedef vec6 search_vec_t;


/*! A basic implementation of the method of Nelder and Mead for minimization of
	an objective function. The implementation follows the description in their
	paper exactly and uses the parameter values described there:
	John A. Nelder and Roger Mead 1965, A Simplex Method for Function
	Minimization, The Computer Journal 7:4, Oxford Academic, p. 308-313,
	https://doi.org/10.1093/comjnl/7.4.308
	\param smallest_value If not nullptr, this float is overwritten by the
		objective value at the returned point.
	\param initialization An initial guess for the location of the minimum.
		This point is used as vertex of the initial simplex. If there are many
		minima, the method will produce different minima dependent on the
		initialization.
	\param objective The objective function with signature
		Float(const search_vec_t&). The given input may be any vector but it is
		legitimate to clamp it or to return large values for invalid inputs.
	\return The location of the smallest value ever observed during the
		optimization.*/
template<class objective_t>
inline search_vec_t minimize_nelder_mead(Float* smallest_value, const search_vec_t& initialization, const objective_t& objective) {
	// The reflection coefficient alpha
	const Float reflection_coeff = 1.0f;
	// The contraction coefficient beta
	const Float contraction_coeff = 0.5f;
	// The expansion coefficient gamma
	const Float expansion_coeff = 2.0f;
	// The shrinking coefficient (implicitly 0.5 in the paper)
	const Float shrink_coeff = 0.5f;
	// Construct the initial simplex using fixed axis-aligned offsets
	search_vec_t simplex[DIMENSIONALITY + 1];
	simplex[0] = initialization;
	for (uint j = 0; j != DIMENSIONALITY; ++j) {
		simplex[j + 1] = initialization;
		simplex[j + 1][j] += 0.05;
	}
	// Evaluate the objective at all simplex vertices. This array is kept up to
	// date at all times.
	Float simplex_value[DIMENSIONALITY + 1];
	for (uint j = 0; j != DIMENSIONALITY + 1; ++j)
		simplex_value[j] = objective(simplex[j]);
	// We also keep track of the best value ever seen in case it is not part of
	// the simplex (which may be the case in presence of failed expansions)
	search_vec_t best = simplex[0];
	Float best_value = simplex_value[0];
	// We use a fixed number of iterations
	for (uint i = 0; i != 100; ++i) {
		// Find indices of minimal and maximal value
		uint min_index = 0, max_index = 0;
		Float min = simplex_value[0], max = simplex_value[0];
		for (uint j = 1; j != DIMENSIONALITY + 1; ++j) {
			min_index = (min > simplex_value[j]) ? j : min_index;
			min = (min > simplex_value[j]) ? simplex_value[j] : min;
			max_index = (max < simplex_value[j]) ? j : max_index;
			max = (max < simplex_value[j]) ? simplex_value[j] : max;
		}

		// Compute the centroid of the facet opposite to the worst vertex
		search_vec_t centroid;
		for (uint j = 0; j != DIMENSIONALITY; ++j)
			centroid[j] = 0.0;
		for (uint j = 0; j != DIMENSIONALITY + 1; ++j)
			if (j != max_index)
				centroid += simplex[j];
		centroid *= 1.0 / DIMENSIONALITY;
		// Compute the reflected point
		search_vec_t reflected = (1.0 + reflection_coeff) * centroid - reflection_coeff * simplex[max_index];
		Float reflected_value = objective(reflected);

		// If the reflected point is better than the best vertex, we expand
		// further in that direction
		if (reflected_value < simplex_value[min_index]) {
			search_vec_t expanded = expansion_coeff * reflected + (1.0 - expansion_coeff) * centroid;
			Float expanded_value = objective(expanded);
			// If the expanded point is also better than the best vertex, we
			// replace the worst vertex by it
			if (expanded_value < simplex_value[min_index]) {
				simplex[max_index] = expanded;
				simplex_value[max_index] = expanded_value;
				// If the reflected value is the best yet, we should not lose
				// it
				if (reflected_value < best_value) {
					best = reflected;
					best_value = reflected_value;
				}
			}
			// Otherwise, we have a failed expansion and use the reflected
			// value instead
			else {
				simplex[max_index] = reflected;
				simplex_value[max_index] = reflected_value;
			}
		}
		else {
			// Check whether replacing the worst vertex by the reflected vertex
			// makes some other vertex the worst
			bool reflected_still_worst = true;
			for (uint j = 0; j != DIMENSIONALITY + 1; ++j)
				reflected_still_worst &= (j == max_index || simplex_value[j] < reflected_value);
			if (reflected_still_worst) {
				// If the reflected point is better than the worst vertex, use
				// it anyway
				if (reflected_value < simplex_value[max_index]) {
					simplex[max_index] = reflected;
					simplex_value[max_index] = reflected_value;
				}
				// Form a contracted point between worst vertex and centroid
				search_vec_t contracted = contraction_coeff * simplex[max_index] + (1.0 - contraction_coeff) * centroid;
				Float contracted_value = objective(contracted);
				// Replace the worst vertex by the contracted point if it makes
				// an improvement
				if (contracted_value < simplex_value[max_index]) {
					simplex[max_index] = contracted;
					simplex_value[max_index] = contracted_value;
				}
				// Otherwise, we shrink all vertices towards the best vertex
				else {
					for (uint j = 0; j != DIMENSIONALITY + 1; ++j) {
						if (j != min_index) {
							simplex[j] = shrink_coeff * simplex[j] + (1.0f - shrink_coeff) * simplex[min_index];
							simplex_value[j] = objective(simplex[j]);
						}
					}
				}
			}
			// At least two vertices are worse than the reflected point, so we
			// replace the worst vertex by it
			else {
				simplex[max_index] = reflected;
				simplex_value[max_index] = reflected_value;
			}
		}
	}
	// The best value ever observed is either a vertex of the simplex, or it
	// has been stored in best_value
	for (uint j = 0; j != DIMENSIONALITY + 1; ++j) {
		if (simplex_value[j] < best_value) {
			best = simplex[j];
			best_value = simplex_value[j];
		}
	}
	if (smallest_value)
		(*smallest_value) = best_value;
	return best;
}
