#  Copyright (C) 2021, Christoph Peters, Karlsruhe Institute of Technology
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.


import numpy as np
import re
import itertools


def compute_assignment_strategy(new_sequence, vertex_count, first_above):
    """
    Helper for print_clipping_code(). It takes a sequence of indices for the
    desired output vertex array and figures out what to write when.
    :param new_sequence: Indices for the original sequence or -1 for the vertex
        on the edge going down or -2 for the vertex on the edge going up.
    :param vertex_count: The vertex count in the original sequence.
    :param first_above: Index of the first vertex in the original sequence that
        is above the horizon.
    :return: An array of tuples (intersect, dst, lhs, rhs). intersect can be
        True to indicate that the intersection for the edge with indices lhs,
        rhs in the current list should be computed and assigned to dst.
        Otherwise lhs will be copied to dst. If the method was unable to find a
        strategy, it returns None.
    """
    # Figure out whether the first vertex is being repeated at the end
    first_repeated = (new_sequence[0] == new_sequence[-1])
    # Enumerate which entries need to be changed
    old_padded = -3 * np.ones(np.maximum(new_sequence.size, vertex_count), dtype=np.int)
    old_padded[0:vertex_count] = np.arange(vertex_count)
    entries_to_change = list()
    for i in range(new_sequence.size):
        if old_padded[i] != new_sequence[i]:
            entries_to_change.append(i)
    # Maintain which entries have been changed to their final value already
    entry_changed = np.zeros_like(old_padded, dtype=np.bool)
    # Compute a few relevant indices
    new_vertex_count = new_sequence.size - (1 if first_repeated else 0)
    count_above = new_vertex_count - 2
    before_first_above = (first_above - 1) if first_above > 0 else (vertex_count - 1)
    last_above = (first_above + count_above - 1) % vertex_count
    after_last_above = (last_above + 1) % vertex_count

    def find_vertex(index):
        """
        :return: An index where the vertex with the given index in the original
            sequence can be found currently. -1 if it has been overwritten.
        """
        if index >= 0 and not entry_changed[index]:
            return index
        for dst, src in enumerate(new_sequence):
            if src == index and entry_changed[dst]:
                return dst
        return -1
    # Iterate over all possible orders in which assignments may be made to find
    # a strategy that works
    strategy = None
    for ordered_entries_to_change in itertools.permutations(entries_to_change):
        entry_changed[...] = False
        strategy = list()
        works = True
        for dst in ordered_entries_to_change:
            src = new_sequence[dst]
            # If the vertex is produced by clipping, we do not want to compute
            # it twice
            if src < 0:
                clipped_index = find_vertex(src)
                if clipped_index >= 0:
                    strategy.append((False, dst, clipped_index, 0))
                    entry_changed[dst] = True
                    continue
            # Figure out where to find the source vertices (if possible)
            if src >= 0:
                srcs = [src]
            elif src == -1:
                srcs = [last_above, after_last_above]
            elif src == -2:
                srcs = [before_first_above, first_above]
            src_indices = [find_vertex(index) for index in srcs]
            works = works and all(np.asarray(src_indices) >= 0)
            if len(src_indices) == 1:
                strategy.append((False, dst, src_indices[0], 0))
            else:
                strategy.append((True, dst, src_indices[0], src_indices[1]))
            entry_changed[dst] = True
        if works:
            break
        else:
            strategy = None
    if strategy is None:
        return None
    return strategy


def get_clipping_code(vertex_count, array_tight, language="glsl", counts_above=None):
    """
    Goes through all possible combinations of vertices being above or below the
    horizon in a convex polygon and returns code to handle these situations.
    :param vertex_count: Number of vertices in the polygon before clipping.
    :param array_tight: If this is False, the output array is allowed to be
        larger than vertex_count + 2, such that the first vertex needs to be
        repeated in any case.
    :param language: "glsl" or "python"
    :param counts_above: Numbers of vertices above the horizon that should be
        handled by the generated code. Defaults to range(vertex_count + 1), i.e.
        all meaningful values.
    :return: A string of code in the requested language.
    """
    code = [""]
    if counts_above is None:
        counts_above = range(vertex_count + 1)

    def write(string, indent=0):
        indentation = "    " if language == "python" else "\t"
        code[0] += (indent * indentation if indent > 0 else "   ") + string
    # For polygons completely below the horizon, we just set the vertex count
    # to zero
    if 0 in counts_above:
        if language == "glsl":
            write("case %4d:   vc = 0;   break;\n" % vertex_count, 1)
        elif language == "python":
            write("if bit_mask == %4d:   vc = 0;\n" % vertex_count, 2)
    # For polygons completely above the horizon, we just need to set the vertex
    # count and repeat the first vertex
    if vertex_count in counts_above:
        above_bitmask = (2 ** vertex_count - 1) * 8 + vertex_count
        if language == "glsl":
            write("case %4d:   vc = %d;   v[%d] = v[0];   break;\n" % (above_bitmask, vertex_count, vertex_count), 1)
        elif language == "python":
            write("if bit_mask == %4d:   vc = %d;   v[%d] = v[0]\n" % (above_bitmask, vertex_count, vertex_count), 2)
    # Other cases are more numerous and more complicated
    for count_above in counts_above:
        if count_above == 0 or count_above == vertex_count:
            continue
        for first_above in range(vertex_count):
            # Mark vertices above the horizon
            mask = np.zeros(vertex_count, dtype=np.int)
            mask[0:count_above] = 1
            mask = np.roll(mask, first_above)
            # Construct the bit mask for the case
            bit_mask = np.sum(mask * 2 ** (np.arange(vertex_count) + 3)) + vertex_count
            if language == "glsl":
                write("case %4d:" % bit_mask, 1)
            else:
                write("if bit_mask == %4d:" % bit_mask, 2)
            # Perform clipping and record the results. -1 means the vertex of
            # the edge going down, -2 is the up vertex
            new_sequence = list()
            for j in range(vertex_count):
                next_vertex = (j + 1) % vertex_count
                if mask[j] == 0 and mask[next_vertex] == 1:
                    new_sequence.append(-2)
                if mask[j] == 1:
                    new_sequence.append(j)
                if mask[j] == 1 and mask[next_vertex] == 0:
                    new_sequence.append(-1)
            new_sequence = np.asarray(new_sequence)
            # Go through all possible cyclic shifts of this sequence to figure
            # out which of them requires the smallest number of move
            # instructions
            new_vertex_count = count_above + 2
            min_operation_count = 1000 + vertex_count
            optimal_sequence = None
            for shift in range(new_vertex_count):
                shifted_sequence = np.roll(new_sequence, shift)
                if not array_tight or shifted_sequence.size < vertex_count + 1:
                    shifted_sequence = np.concatenate([shifted_sequence, [shifted_sequence[0]]])
                strategy = compute_assignment_strategy(shifted_sequence, vertex_count, first_above)
                if strategy is not None:
                    operation_count = len(strategy)
                    if min_operation_count > operation_count:
                        min_operation_count = operation_count
                        optimal_sequence = shifted_sequence
            if optimal_sequence is None:
                print("Failed to find a strategy for any cyclic shift of sequence %s." % str(new_sequence))
                continue
            # Write the vertex count
            write("vc = %d;" % new_vertex_count)
            # Generate code for the optimal assignment strategy
            strategy = compute_assignment_strategy(optimal_sequence, vertex_count, first_above)
            for intersect, dst, lhs, rhs in strategy:
                if language == "glsl":
                    if intersect:
                        write("v[%d] = iz0(v[%d], v[%d]);" % (dst, lhs, rhs))
                    else:
                        write("v[%d] = v[%d];" % (dst, lhs))
                elif language == "python":
                    if intersect:
                        write("v[%d, :] = iz0(v[%d, :], v[%d, :]);" % (dst, lhs, rhs))
                    else:
                        write("v[%d, :] = v[%d, :];" % (dst, lhs))
            if language == "glsl":
                write("break;\n")
            elif language == "python":
                write("\n")
            print("%4d" % bit_mask, end=" ")
    # Discard trailing semicolons in Python
    if language == "python":
        code[0] = code[0].replace(";   \n", "\n")
    return code[0]


if __name__ == "__main__":
    for language in ["python", "glsl"]:
        code = ""
        for count in range(3, 8):
            if language == "glsl":
                code += "#if MIN_POLYGON_VERTEX_COUNT_BEFORE_CLIPPING <= %d && MAX_POLYGON_VERTEX_COUNT >= %d\n" % (count, count + 1)
                code += get_clipping_code(count, False, language, list(range(count - 1)) + [count])
                code += "#endif\n"
                code += "#if MIN_POLYGON_VERTEX_COUNT_BEFORE_CLIPPING <= %d && MAX_POLYGON_VERTEX_COUNT == %d\n" % (count, count + 1)
                code += get_clipping_code(count, True, language, [count - 1])
                code += "#elif MIN_POLYGON_VERTEX_COUNT_BEFORE_CLIPPING <= %d && MAX_POLYGON_VERTEX_COUNT > %d\n" % (count, count + 1)
                code += get_clipping_code(count, False, language, [count - 1])
                code += "#endif\n"
            elif language == "python":
                code += "    if max_vertex_count >= %d:\n" % (count + 1)
                code += get_clipping_code(count, False, language, list(range(count - 1)) + [count])
                code += "    if max_vertex_count == %d:\n" % (count + 1)
                code += get_clipping_code(count, True, language, [count - 1])
                code += "    elif max_vertex_count > %d:\n" % (count + 1)
                code += get_clipping_code(count, False, language, [count - 1])
        target_path = "../vulkan_renderer/src/shaders/polygon_clipping.glsl" if language == "glsl" else "polygon_clipping.py"
        with open(target_path, "r") as file:
            unmodified_code = file.read()
        modified_code = re.sub(r"((//|#) AUTOGENERATED PART BEGIN\s).*(\s(//|#) AUTOGENERATED PART END)", r"\1%s\3" % code, unmodified_code, flags=re.DOTALL)
        with open(target_path, "w") as file:
            file.write(modified_code)
        print("Done.")
