#  Copyright (C) 2021, Christoph Peters, Karlsruhe Institute of Technology
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.


import re
import numpy as np
from matplotlib import pyplot
from os import path
import json
from imageio import imread, imwrite
# import imageio
# imageio.plugins.freeimage.download()
from generate_graphics import list_experiments, compute_error


def make_readable_name(name):
    """Turns a string separated by underscores into something more readable."""
    replacements = [
        ("_", " "),
        (" mis", " MIS"),
        ("mis ", "MIS "),
        ("ggx", "VNDF sampling (Dong)"),
        ("arvo", "(Arvo)"),
        ("turk", "(Turk)"),
        ("urena", "(Urena)"),
        (" hart", " (Hart et al.)"),
        (" veach", " (Veach)"),
        (" ours", " (ours)"),
        ("spp", " spp"),
    ]
    readable_name = name
    for old, new in replacements:
        readable_name = readable_name.replace(old, new)
    readable_name = readable_name[:1].upper() + readable_name[1:]
    return readable_name


def png_to_webp(png_path, webp_path):
    """
    Converts a PNG at the given path to a lossless WEBP at the given path.
    """
    imwrite(webp_path, imread(png_path), flags=0x100)


def prepare_scene(scene_name, skip_name=lambda name: False):
    """
    Gathers images from the experiments directory belonging to the scene with
    the given name. It converts all of them to WEBP images in the directory for
    the HTML viewer. Additionally, it gathers meta data. Images are skipped if
    the given function returns True for their name.
    :return: A dictionary mapping image names to dictionaries of meta data and
        the width and height of the reference image.
    """
    # List relevant images
    png_experiments = list_experiments("%s_.*\.png" % scene_name)
    hdr_experiments = list_experiments("%s_.*\.hdr" % scene_name)
    references = [file_path for file_path, name in hdr_experiments if name == scene_name + "_reference.hdr"]
    if len(references) != 1:
        print("There are %d references among the experiments for scene %s. Need exactly one." % (len(references), scene_name))
        return
    reference = imread(references[0])
    # Process them one by one
    image_dict = dict()
    for file_path, name in png_experiments:
        raw_name = name[0:len(name) - 4]
        if skip_name(file_path):
            continue
        # There must be a match HDR version of each LDR image
        hdr_paths = [hdr_path for hdr_path, hdr_name in hdr_experiments if hdr_name == raw_name + ".hdr"]
        if len(hdr_paths) != 1:
            print("Found %d HDR versions of %s, but need exactly one." % (len(hdr_paths), raw_name))
            continue
        hdr_image = imread(hdr_paths[0])
        # Compute the error
        rmse = compute_error(hdr_image, reference)
        # Extract the sample count and frame time
        match = re.search(r"_([0-9]+)spp_([0-9]+\.[0-9]+)\.png", file_path)
        sample_count = int(match.group(1))
        frame_time = float(match.group(2))
        # Convert to WEBP (if the file does not exist yet)
        webp_path = "../html_viewer/img/%s.webp" % raw_name
        if not path.exists(webp_path):
            png_to_webp(file_path, webp_path)
        # Store meta data
        readable_name = make_readable_name(raw_name.replace(scene_name + "_", ""))
        image_dict[raw_name] = dict(name=readable_name, rmse=rmse, sample_count=sample_count, frame_time=frame_time)
    # Create an RMSE plot
    figure = pyplot.figure(figsize=(8.0, 4.0))
    timings = [image["frame_time"] for image in image_dict.values() if image["rmse"] != 0.0]
    rmses = [image["rmse"] for image in image_dict.values() if image["rmse"] != 0.0]
    x_lim = np.asarray([np.min(timings) * 0.9, np.max(timings) * 1.1])
    pyplot.subplot(xlabel="Total frame time / ms", ylabel="RMSE", xscale="log", yscale="log")
    pyplot.xlim(x_lim)
    pyplot.ylim(0.75 * np.min(rmses), 1.1 * np.max(rmses))
    for key, image in image_dict.items():
        if image["rmse"] == 0.0:
            continue
        extrapolated_rmse = image["rmse"] / np.sqrt(x_lim / image["frame_time"])
        pyplot.plot(x_lim, extrapolated_rmse, zorder=-200.0, color=3 * [0.9], linewidth=1.0)
        rank = np.count_nonzero(rmses < image["rmse"])
        scatter_color = "C%d" % rank
        pyplot.scatter(image["frame_time"], image["rmse"], color=scatter_color)
        pyplot.annotate(image["name"], (image["frame_time"], image["rmse"]),
                        (image["frame_time"], np.logspace(np.log10(np.min(rmses)), np.log10(np.max(rmses)), len(image_dict) - 1)[rank]),
                        arrowprops=dict(arrowstyle="-", color=scatter_color), ha="center", va="top")
    figure.savefig("../html_viewer/img/%s_error.png" % scene_name, dpi=300.0, bbox_inches="tight")
    return image_dict, reference.shape[1], reference.shape[0]


if __name__ == "__main__":
    scene_dict = dict()
    scene_descriptions = dict(
        arcade="An arcade with a heptagonal light (i.e. seven vertices) mounted to the wall. The floor and the poster show prominent specular highlights that benefit from our LTC importance sampling. Projected solid angle sampling helps near the light. Scene created by Nicholas Hull for Falcor by NVIDIA.",
        attic="A scene with illumination from a light probe through a large pentagonal portal. The floor has low roughness and benefits strongly from specular importance sampling, other parts are more diffuse.",
        bistro_small_polygon="A wall lit by a 1 square centimeter light source 40 meters away. A bollard close to the light casts a soft shadow, the lantern casts a harder shadow. Some techniques for sampling the solid angle of the light fail completely due to numerical issues, others occasionally produce directions that miss the light.",
        bistro_tiny_polygon="A wall lit by a 1 square millimeter light source 40 meters away. A bollard close to the light casts a slighlty soft shadow, the lantern casts a hard shadow. All techniques except area sampling run into numerical issues but to a different extent.",
        cornell_box="A Cornell box with high roughness. It uses the Frostbite BRDF, i.e. it is not just Lambertian diffuse. Nonetheless, sampling proportional to projected solid angle is highly effective outside of penumbrae. Solid angle sampling has high variance on the left and right wall, area sampling is bad near the light source.",
        living_room="A living room with a square 40 cm light mounted to the ceiling. Area sampling still has considerable drawbacks but the cost of sampling strategies more sophisticated than solid angle sampling is harder to justify. The RMSE is dominated by the variance on the bright ceiling but elsewhere the image quality barely changes. Scene created by Blend Swap user Jay-Artist.",
        mis_plane="A diffuse and a glossy plane, lit by a large rectangular light that is half-shadowed. It is an interesting test case for different MIS heuristics combining diffuse and specular samples. The balance and power heuristic struggle in fully lit regions, our weighted balance heuristic is suboptimal in the penumbra. Clamped optimal MIS with v=0.5 is a reasonable middle ground.",
        shadowed_plane="A plane with low roughness, a large light and partial occlusion provokes noticeable bias in our biased variant of projected solid angle sampling. These images use our clamped optimal MIS.",
    )

    def skip_image_name(name):
        return name.find("_baseline") >= 0 or name.find("_tilted") >= 0
    for scene_name, description in scene_descriptions.items():
        image_dict, width, height = prepare_scene(scene_name, skip_image_name)
        images = list(image_dict.items())
        images = sorted(images, key=lambda image: -image[1]["rmse"])
        images = [image[0] for image in images]
        scene_dict[scene_name] = dict(images=images, image_dict=image_dict, name=make_readable_name(scene_name),
                                      description=description, width=width, height=height)
    with open("../html_viewer/images.js", "w") as json_file:
        json_file.write("scene_dict = " + json.dumps(scene_dict, indent="\t") + ";")
        scenes = sorted(scene_dict.keys())
        json_file.write("scenes = " + json.dumps(scenes) + ";")
