#  Copyright (C) 2021, Christoph Peters, Karlsruhe Institute of Technology
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.


import numpy as np
from matplotlib import pyplot
from polygon_clipping import clip_polygon


def positive_atan(tangent):
    """
    :return: The arctangent as angle between 0 and pi.
    """
    return np.arctan(tangent) + np.where(tangent < 0.0, np.pi, 0.0)


def is_inner(ellipse):
    """
    :return: True if the given ellipse is tagged as inner ellipse.
    """
    return np.signbit(ellipse[0])


def dot(lhs, rhs):
    """
    Dot product for the first dimension of two arrays of identical shape.
    """
    return np.einsum("i..., i... -> ...", lhs, rhs)


def outer(lhs, rhs):
    """
    Outer product for two vectors of shapes (row_count, ...) and
    (column_count, ...). The result has shape (row_count, column_count, ...).
    """
    return np.einsum("i..., j... -> ij...", lhs, rhs)


def ellipse_transform(ellipse, point):
    """
    :param ellipse: See ellipse_from_edge()
    :param point: The point to be transformed (shape (2, ...)).
    :return: point transformed using the matrix that characterizes the ellipse.
    """
    return point + ellipse * dot(ellipse, point)


def rotate_90(vector):
    """Rotates the given vector (shape (2, ...) 90 degrees counterclockwise."""
    return np.stack([-vector[1], vector[0]])


def ellipse_area_in_sector(ellipse, sector_begin, sector_end):
    """
    :param ellipse: See ellipse_from_edge().
    :param sector_begin: Direction vector for the start of the sector
        (shape (2, ...)).
    :param sector_end: End of the sector (less than 180 degrees
        counterclockwise of its beginning.
    :return: The area in the intersection of the ellipse with the given sector.
    """
    rsqrt_det = 1.0 / np.sqrt(1.0 + dot(ellipse, ellipse))
    sector_det = np.maximum(0.0, -dot(sector_begin, rotate_90(sector_end)))
    ellipse_dot = rsqrt_det * dot(sector_begin, ellipse_transform(ellipse, sector_end))
    area = 0.5 * rsqrt_det * positive_atan(sector_det / ellipse_dot)
    return np.where(area > 0.0, area, 0.0)


def ellipse_from_edge(vertex_0, vertex_1):
    """
    Takes the great circle through the given two vertices and projects it down
    to the xy-plane. The resulting ellipse is expressed in terms of a vector u
    such that dot(q, q) + dot(u, q) ** 2 = 1 for all points q on the ellipse.
    :param vertex_0: Non-zero direction vector from the shading point to the
        first vertex along the edge (shape (3, ...)).
    :param vertex_1: Same as vertex_0 for the other vertex.
    :return: The vector with shape (2, ...). The sign bit of the first entry
        indicates whether the edge runs clockwise (0) or counterclockwise (1).
        Iff the first entry is infinity, the great circle runs through the
        zenith.
    """
    # We should be using Kahan's algorithm here but NumPy does not expose
    # fused-multiply add, so we don't
    normal = np.cross(vertex_0, vertex_1, axis=0)
    scaling = 1.0 / normal[2]
    scaling = np.where(is_inner(normal[0:2]), -scaling, scaling)
    ellipse = normal[0:2] * scaling
    ellipse[0] = np.where(normal[2] == 0.0, np.infty, ellipse[0])
    return ellipse


def solve_homogeneous_quadratic(quadratic):
    """
    :param quadratic: A matrix (shape (2, 2, ...)).
    :return: A vector (shape (2, ...)) such that
        dot(vector, matrix @ vector) == 0.
    """
    coeff_xy = 0.5 * (quadratic[0, 1] + quadratic[1, 0])
    discriminant = coeff_xy * coeff_xy - quadratic[0, 0] * quadratic[1, 1]
    scaled_solution = np.abs(coeff_xy) + np.sqrt(np.maximum(0.0, discriminant))
    return np.where(coeff_xy >= 0.0,
                    np.stack([scaled_solution, -quadratic[0, 0]]),
                    np.stack([quadratic[1, 1], scaled_solution]))


def sample_sector_between_ellipses(random_numbers, target_area, inner_ellipse, outer_ellipse, begin, end, compute_backward_error, iteration_count=2):
    """
    Samples the area between two ellipses uniformly within the given sector.
    All inputs and outputs have shape (2, ...), except target_area and the
    error, which has shape (...).
    :param random_numbers: Uniform random numbers on [0, 1)^2.
    :param target_area: random_numbers[0] times the size of the sampled area.
    :param inner_ellipse: See ellipse_from_edge().
    :param outer_ellipse: See ellipse_from_edge().
    :param begin: A direction vector for the beginning of the sector.
    :param end: Less than 180 degrees counterclockwise from the beginning.
    :param compute_backward_error: Pass True to get the backward error in terms
        of target_area instead of a sample.
    :param iteration_count: The default iteration count.
    :return: A uniform sample within the specified area (up to high accuracy) or
        the error thereof.
    """
    target_area = np.asarray(target_area)
    # Construct the half-vector to halve the sector
    begin = begin / np.linalg.norm(begin, axis=0)
    end = end / np.linalg.norm(end, axis=0)
    middle = begin + end
    # Compute intersections of ellipses with sector boundaries
    base_shape = list(target_area.shape)
    factors = np.zeros([2, 3] + base_shape)
    for l, ellipse in enumerate([inner_ellipse, outer_ellipse]):
        for j, sector_dir in enumerate([begin, middle, end]):
            factors[l, j] = 1.0 / np.sqrt(dot(sector_dir, sector_dir) + dot(ellipse, sector_dir) ** 2)
    # Compute areas of quads (except for a common factor)
    quad_area_0 = factors[1, 0] * factors[1, 1] - factors[0, 0] * factors[0, 1]
    quad_area_1 = factors[1, 1] * factors[1, 2] - factors[0, 1] * factors[0, 2]
    # Select a quad and move its attributes to index 2
    target_quad_area = (1.0 - random_numbers[0]) * -quad_area_0 + random_numbers[0] * quad_area_1
    swap = target_quad_area <= 0.0
    end = np.where(swap, begin, end)
    factors[:, 2] = np.where(swap, factors[:, 0], factors[:, 2])
    target_quad_area += np.where(swap, quad_area_0, -quad_area_1)
    target_quad_area *= np.abs(dot(middle, rotate_90(end)))
    # Construct line equations for the edges of the quad
    edge_normals = np.zeros([2, 2] + base_shape)
    edge_offsets = np.zeros([2] + base_shape)
    for l, ellipse in enumerate([inner_ellipse, outer_ellipse]):
        edge_normals[l] = factors[l, 1] * middle + factors[l, 2] * end
        edge_normals[l] = ellipse_transform(ellipse, edge_normals[l])
        edge_offsets[l] = factors[l, 1] * dot(edge_normals[l], middle)
    quadratic = outer((factors[1, 2] * edge_offsets[1]) * rotate_90(end), edge_normals[0])
    quadratic -= outer((factors[0, 2] * edge_offsets[0]) * rotate_90(end) + target_quad_area * edge_normals[0], edge_normals[1])
    initial_dir = solve_homogeneous_quadratic(quadratic)
    # Run iterations
    current_dir = np.copy(initial_dir)
    inner_rsqrt_det = 1.0 / np.sqrt(1.0 + dot(inner_ellipse, inner_ellipse))
    outer_rsqrt_det = 1.0 / np.sqrt(1.0 + dot(outer_ellipse, outer_ellipse))
    for n in range((iteration_count + 1) if compute_backward_error else iteration_count):
        # Avoid under- or overflow
        current_dir /= np.abs(current_dir[0]) + np.abs(current_dir[1])
        # Flip the sign if the direction is outside the sector
        current_dir = np.where(dot(middle, current_dir) < 0.0, -current_dir, current_dir)
        # Compute the error in the target function
        inner_dir = ellipse_transform(inner_ellipse, current_dir)
        inner_area = 0.5 * inner_rsqrt_det * positive_atan(np.maximum(0.0, -dot(begin, rotate_90(current_dir))) / (inner_rsqrt_det * dot(begin, inner_dir)))
        outer_dir = ellipse_transform(outer_ellipse, current_dir)
        outer_area = 0.5 * outer_rsqrt_det * positive_atan(np.maximum(0.0, -dot(begin, rotate_90(current_dir))) / (outer_rsqrt_det * dot(begin, outer_dir)))
        error = target_area - outer_area + inner_area
        if compute_backward_error and n == iteration_count:
            return error
        # Construct a homogeneous quadratic and solve it
        quadratic = outer(rotate_90(current_dir), inner_dir - outer_dir) - outer(2.0 * error * inner_dir, outer_dir)
        current_dir = solve_homogeneous_quadratic(quadratic)
    # If we should not have iterated, we revert the result now
    no_iterations = (random_numbers[0] < 1.0e-5) | (random_numbers[0] > 1.0 - 1.0e-5)
    current_dir = np.where(no_iterations, initial_dir, current_dir)
    # Flip the sign if the direction is outside of the sector
    current_dir = np.where(dot(middle, current_dir) < 0.0, -current_dir, current_dir)
    # Sample an appropriate length
    inner_scale = 1.0 / (dot(current_dir, current_dir) + dot(inner_ellipse, current_dir) ** 2)
    outer_scale = 1.0 / (dot(current_dir, current_dir) + dot(outer_ellipse, current_dir) ** 2)
    scale = (1.0 - random_numbers[1]) * inner_scale + random_numbers[1] * outer_scale
    return np.sqrt(scale) * current_dir


class Polygon:
    """
    Gathers attributes needed to sample a polygon proportional to projected
    solid angle.
    """

    def __init__(self, vertices):
        """
        Prepares sampling of the convex polygon with the given vertices.
        :param vertices: An array of shape (vertex_count, 3) providing the
            vertices of the polygon. They must be in a coordinate frame where
            the shading point is in the origin and the normal is aligned with
            the z-axis. The winding as seen from the origin must be clockwise.
            No three vertices should be collinear. vertex_count must be at most
            seven.
        """
        # Clip the polygon
        new_vertex_count, clipped_vertices = clip_polygon(vertices, vertices.shape[0] + 2)
        vertices = clipped_vertices[0:new_vertex_count + 1]
        self.vertex_count = vertices.shape[0] - 1
        # Associate vertices with ellipses
        self.ellipses = np.zeros((self.vertex_count, 2))
        self.ellipses[0] = ellipse_from_edge(vertices[0], vertices[1])
        previous_ellipse = self.ellipses[0]
        self.inner_ellipse_0 = np.asarray([1.0, 0.0])
        for j in range(1, self.vertex_count):
            current_ellipse = ellipse_from_edge(vertices[j], vertices[j + 1])
            self.ellipses[j] = previous_ellipse if is_inner(current_ellipse) else current_ellipse
            if is_inner(previous_ellipse) and not is_inner(current_ellipse):
                self.inner_ellipse_0 = np.copy(previous_ellipse)
            previous_ellipse = current_ellipse
        # One more time to close the edge loop
        current_ellipse = self.ellipses[0]
        self.ellipses[0] = previous_ellipse if is_inner(current_ellipse) else current_ellipse
        if is_inner(previous_ellipse) and not is_inner(current_ellipse):
            self.inner_ellipse_0 = np.copy(previous_ellipse)
        # From this point onward, we only need x- and y-coordinates of vertices
        self.vertices = vertices[:, 0:2]
        # Compute the projected solid angle sector by sector
        self.projected_solid_angle = 0.0
        self.sector_projected_solid_angles = np.zeros(self.vertex_count)
        if self.central_case():
            for j in range(self.vertex_count):
                area = ellipse_area_in_sector(self.ellipses[j], self.vertices[j], self.vertices[j + 1])
                self.sector_projected_solid_angles[j] = area
                self.projected_solid_angle += area
        else:
            self.sort_sectors()
            inner_ellipse = self.inner_ellipse_0
            for j in range(self.vertex_count - 1):
                if j > 0 and is_inner(self.ellipses[j]):
                    inner_ellipse = self.ellipses[j]
                if j == 0 or not is_inner(self.ellipses[j]):
                    outer_ellipse = self.ellipses[j]
                area = ellipse_area_in_sector(outer_ellipse, self.vertices[j], self.vertices[j + 1])
                area -= ellipse_area_in_sector(inner_ellipse, self.vertices[j], self.vertices[j + 1])
                area = np.maximum(0.0, area)
                self.sector_projected_solid_angles[j] = area
                self.projected_solid_angle += area

    def central_case(self):
        """
        :return: True iff the central case is present, i.e. the zenith is part
            of the solid angle of the polygon.
        """
        return not is_inner(self.inner_ellipse_0)

    def sort_sectors(self):
        """
        Sorts vertices counter clockwise with appropriate tie breaking based on
        the associated ellipses.
        """
        # Optimal sorting networks for bitonic sequences of lengths up to eight
        # elements. Note that the last three comparisons are shared for all
        # networks with more than three elements. That enables elimination of
        # redundant code (which we do not do here).
        sorting_networks = [
            [], [], [(0, 1)],
            [(1, 2), (0, 2), (0, 1)], [(1, 3), (0, 2), (2, 3), (0, 1)],
            [(2, 4), (1, 3), (1, 2), (0, 3), (3, 4), (0, 2), (2, 3), (0, 1)],
            [(3, 5), (2, 4), (1, 5), (0, 4), (4, 5), (1, 3), (0, 2), (2, 3), (0, 1)],
            [(2, 5), (1, 6), (5, 6), (3, 4), (0, 4), (4, 6), (1, 3), (3, 5), (4, 5), (0, 2), (2, 3), (0, 1)],
            [(2, 6), (3, 7), (1, 5), (0, 4), (4, 6), (5, 7), (6, 7), (4, 5), (1, 3), (0, 2), (2, 3), (0, 1)]
        ]
        for lhs, rhs in sorting_networks[self.vertex_count]:
            # We should use Kahan's algorithm here but NumpPy does not expose
            # fused multiply-add, so we don't
            vertex_det = dot(rotate_90(self.vertices[lhs]), self.vertices[rhs])
            swap = (vertex_det < 0.0 or (vertex_det == 0.0 and np.isinf(self.vertices[rhs, 0])))
            lhs_copy = np.copy(self.vertices[lhs])
            self.vertices[lhs] = self.vertices[rhs] if swap else lhs_copy
            self.vertices[rhs] = lhs_copy if swap else self.vertices[rhs]
            lhs_copy = np.copy(self.ellipses[lhs])
            self.ellipses[lhs] = self.ellipses[rhs] if swap else lhs_copy
            self.ellipses[rhs] = lhs_copy if swap else self.ellipses[rhs]

    def sample(self, random_numbers, compute_backward_error=False, iteration_count=2):
        """
        Takes one or more samples from the polygon. They are distributed
        uniformly with respect to projected solid angle.
        :param random_numbers: Uniform random numbers in [0,1)^2 (shape
            (2, ...)).
        :param compute_backward_error: Pass True if you do not care about the
            actual samples but only about the error in random_numbers[0] that
            suffices to explain their deviation from an exact result. Zero for
            the central case.
        :param iteration_count: The default iteration count.
        :return: Normalized vectors in the upper hemisphere (shape (3, ...)) or
            an array with errors (shape (...)).
        """
        random_numbers = np.asarray(random_numbers)
        target_area = self.projected_solid_angle * random_numbers[0]
        base_shape = list(target_area.shape)
        if self.central_case():
            if compute_backward_error:
                return np.zeros_like(target_area)
            # The loop may break after a different number of iterations for
            # each random number, so we set a write mask to False when it
            # breaks
            write_mask = np.ones(base_shape, dtype=np.bool)
            outer_ellipse = np.zeros([2] + base_shape)
            sector_begin = np.zeros([2] + base_shape)
            # Select a sector
            for j in range(self.vertex_count):
                if j > 0:
                    target_area[write_mask] -= self.sector_projected_solid_angles[j - 1]
                for k in range(2):
                    outer_ellipse[k][write_mask] = self.ellipses[j, k]
                    sector_begin[k][write_mask] = self.vertices[j, k]
                write_mask &= (target_area >= self.sector_projected_solid_angles[j])
            # Sample the direction within the sector
            sqrt_det = np.sqrt(1.0 + dot(outer_ellipse, outer_ellipse))
            scaled_area = 2.0 * sqrt_det * target_area
            sampled_dir = (np.cos(scaled_area) * sqrt_det) * sector_begin
            sampled_dir += np.sin(scaled_area) * rotate_90(ellipse_transform(outer_ellipse, sector_begin))
            # Sample its inclination
            squared_normalization = 1.0 / dot(sampled_dir, ellipse_transform(outer_ellipse, sampled_dir))
            sampled_dir = np.sqrt(random_numbers[1] * squared_normalization) * sampled_dir
        else:
            # Same construct
            write_mask = np.ones(base_shape, dtype=np.bool)
            inner_ellipse, outer_ellipse, sector_begin, sector_end = [np.zeros([2] + base_shape) for j in range(4)]
            sector_area = np.zeros_like(target_area)
            inner_ellipse[0] = self.inner_ellipse_0[0]
            inner_ellipse[1] = self.inner_ellipse_0[1]
            # Select a sector
            for j in range(self.vertex_count - 1):
                ellipse = self.ellipses[j]
                if j == 0:
                    outer_ellipse[0] = ellipse[0]
                    outer_ellipse[1] = ellipse[1]
                else:
                    target_area[write_mask] -= self.sector_projected_solid_angles[j - 1]
                    for k in range(2):
                        inner_ellipse[k][write_mask & is_inner(ellipse)] = ellipse[k]
                        outer_ellipse[k][write_mask & ~is_inner(ellipse)] = ellipse[k]
                for k in range(2):
                    sector_begin[k][write_mask] = self.vertices[j, k]
                    sector_end[k][write_mask] = self.vertices[j + 1, k]
                sector_area[write_mask] = self.sector_projected_solid_angles[j]
                write_mask &= (target_area >= sector_area)
            # Sample it
            random_numbers = np.copy(random_numbers)
            random_numbers[0] = target_area / sector_area
            result = sample_sector_between_ellipses(random_numbers, target_area, inner_ellipse, outer_ellipse,
                                                    sector_begin, sector_end, compute_backward_error, iteration_count)
            if compute_backward_error:
                return result / self.projected_solid_angle
            else:
                sampled_dir = result
        sampled_dir_z = np.sqrt(np.maximum(0.0, 1.0 - dot(sampled_dir, sampled_dir)))
        return np.concatenate([sampled_dir, sampled_dir_z[np.newaxis]], axis=0)


if __name__ == "__main__":
    # The initial test polygon is equal-sided
    polygon_vertex_count = 4
    angles = np.linspace(-np.pi, np.pi, polygon_vertex_count + 1)[1:]
    polygon_vertices = np.vstack([1.1 + np.cos(angles), 0.3 + np.sin(angles), np.ones_like(angles)]).T
    # This normalization is only needed for the sake of the user-interface
    polygon_vertices /= np.linalg.norm(polygon_vertices, axis=1)[:, np.newaxis]
    # Prepare the interactive plot
    figure = pyplot.figure(figsize=(12.0, 6.0))
    error_axes = pyplot.subplot(1, 2, 2, title="Iteration error", xlabel="Random number used to sample the azimuth", ylabel=r"$\log_{10}$(abs(error in random number))")
    error_plot = pyplot.plot([0.0], [0.0])[0]
    pyplot.xlim(0.0, 1.0)
    pyplot.ylim(-20, 0.0)
    pyplot.grid()
    polygon_axes = pyplot.subplot(1, 2, 1, aspect=1.0)
    # Load blue noise points for sampling
    point_file = "blue_noise_de_goes.dat"
    blue_noise_points = np.fromfile(point_file, sep=" ").reshape((-1, 2)).T
    # Sort them by x-coordinate for the sake of error plots
    blue_noise_points = blue_noise_points[:, np.argsort(blue_noise_points[0])]
    # A few arrays for plotting
    lerp_factor = np.linspace(0.0, 1.0, 1001)
    phase = np.linspace(-np.pi, np.pi, 1001)

    def plot_polygon(vertices):
        pyplot.gca().clear()
        pyplot.title("Click to move closest vertex")
        # Plot edges of the spherical polygon
        for j in range(vertices.shape[0]):
            vertex, next_vertex = vertices[j], vertices[(j + 1) % vertices.shape[0]]
            edge = (1.0 - lerp_factor) * vertex[:, np.newaxis] + lerp_factor * next_vertex[:, np.newaxis]
            edge /= np.linalg.norm(edge, axis=0)
            pyplot.plot(edge[0], edge[1], color="k")
        # Plot the horizon
        pyplot.plot(np.cos(phase), np.sin(phase), color=3 * [0.7])
        # If the winding is wrong we fix it. In a renderer that would be done
        # based on the plane equation of the polygon.
        if np.linalg.det(vertices[:3]) < 0.0:
            vertices = vertices[::-1]
        # Non-convex polygons are prohibited
        triangles = np.stack([vertices, np.roll(vertices, -1, 0), np.roll(vertices, -2, 0)], axis=1)
        determinants = np.linalg.det(triangles)
        if np.any(determinants < 0.0):
            pyplot.text(0.0, 0.0, "Non-convex", ha="center", va="center")
            figure.canvas.draw()
            return
        # Take samples and plot them
        polygon = Polygon(vertices)
        samples = polygon.sample(blue_noise_points)
        pyplot.scatter(samples[0], samples[1], color="C0", s=1.8)
        # Compute and plot the error
        backward_error = polygon.sample(blue_noise_points, True)
        error_plot.set_data(blue_noise_points[0], np.log10(np.maximum(1.0e-30, np.abs(backward_error))))
        figure.canvas.draw()

    def on_click(event):
        global polygon_vertices
        if event.xdata is not None:
            point = np.asarray([event.xdata, event.ydata, 0.0])
            if np.linalg.norm(point) >= 1.0:
                point /= np.sum(point * point)
                point[2] = -np.sqrt(1.0 - np.sum(point * point))
            else:
                point[2] = np.sqrt(1.0 - np.sum(point * point))
            distances = np.linalg.norm(polygon_vertices - point[np.newaxis, :], axis=1)
            index = np.argmin(distances)
            polygon_vertices[index, :] = point
            plot_polygon(polygon_vertices)

    plot_polygon(polygon_vertices)
    figure.canvas.mpl_connect("button_press_event", on_click)
    pyplot.show()
