#  Copyright (C) 2021, Christoph Peters, Karlsruhe Institute of Technology
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.


import numpy as np
import imageio
from matplotlib import colors


def get_image_rectangle(image, left_top, right_bottom):
    """Returns a slice from the given image for the given rectangle. If the 
       rectangle is out of bounds, it is clipped.
    """
    left_top = np.maximum(0, left_top)
    right_bottom = np.minimum(image.shape[1::-1], right_bottom)
    return image[left_top[1]:right_bottom[1], left_top[0]:right_bottom[0], ...]


class Magnifier:
    """Stores a specification of a magnified inset.
    """

    def __init__(self, magnifier_left_top, magnifier_width_height, inset_left_top, magnification, border_color,
                 magnifier_border_width=8, inset_border_width=None):
        """Initializes with the given attributes. inset_border_width defaults to 
           magnifier_border_width. All length units are pixels and only integers 
           are allowed, 2D quantities have to be iterables of length 2, 
           magnification must be an integer. The color can be any matplotlib 
           color specification.
        """
        self.magnifier_left_top = np.asarray(magnifier_left_top)
        self.magnifier_width_height = np.asarray(magnifier_width_height)
        self.inset_left_top = np.asarray(inset_left_top)
        self.magnification = int(magnification)
        self.border_color = np.asarray(255.0 * np.asarray(colors.to_rgb(border_color)), dtype=np.uint8)
        self.magnifier_border_width = int(magnifier_border_width)
        self.inset_border_width = self.magnifier_border_width if inset_border_width is None else int(inset_border_width)

    def draw_magnifier(self, image):
        """Draws the border of the magnifier into the given image (in 
           place).
        """
        if self.magnifier_border_width <= 0:
            return
        # Copy the parts inside the magnifier
        inner_rectangle = get_image_rectangle(image, self.magnifier_left_top,
                                              self.magnifier_left_top + self.magnifier_width_height)
        inner_rectangle_copy = np.copy(inner_rectangle)
        # Draw a filled rectangle
        outer_rectangle = get_image_rectangle(image, self.magnifier_left_top - self.magnifier_border_width,
                                              self.magnifier_left_top + self.magnifier_width_height
                                              + self.magnifier_border_width)
        for j in range(3):
            outer_rectangle[:, :, j] = self.border_color[j]
        # Copy the original contents back in
        inner_rectangle[...] = inner_rectangle_copy

    def draw_inset(self, image):
        """Draws the magnified inset and its border into the given image (in 
           place.)
        """
        source_rectangle = np.copy(
            get_image_rectangle(image, self.magnifier_left_top, self.magnifier_left_top + self.magnifier_width_height))
        # Draw a filled rectangle
        outer_rectangle = get_image_rectangle(image, self.inset_left_top - self.inset_border_width,
                                              self.inset_left_top + self.magnification * self.magnifier_width_height
                                              + self.inset_border_width)
        for j in range(3):
            outer_rectangle[:, :, j] = self.border_color[j]
        # Draw the magnified parts of the image
        inner_rectangle = get_image_rectangle(image, self.inset_left_top,
                                              self.inset_left_top + self.magnification * self.magnifier_width_height)
        inner_rectangle[...] = source_rectangle.repeat(self.magnification, 0).repeat(self.magnification, 1)


class Screenshot:
    """Represents a single RGB image and supports drawing magnified insets to 
       it.
    """

    def __init__(self, source_file_or_array):
        """Loads the image in the given file and initializes this object with 
           no magnifiers. You can also pass an array of shape (height,width,3) 
           with dtype np.uint8.
        """
        if isinstance(source_file_or_array, str):
            self.image = imageio.imread(source_file_or_array)
        else:
            self.image = np.copy(source_file_or_array)
        self.magnifier_list = list()
        self.width = self.image.shape[1]
        self.height = self.image.shape[0]

    def add_magnifier(self, *argument_list):
        """
        Forwards to Magnifier.__init__() and adds this magnifier to the
        screenshot.
        """
        self.magnifier_list.append(Magnifier(*argument_list))

    def add(self, magnifier):
        """Adds the given Magnifier to the screenshot."""
        self.magnifier_list.append(magnifier)

    def get_image_with_magnified_insets(self):
        """Returns a copy of the held image with all magnifiers applied.
        """
        result = np.copy(self.image)
        for magnifier in self.magnifier_list:
            magnifier.draw_magnifier(result)
        for magnifier in self.magnifier_list:
            magnifier.draw_inset(result)
        return result

    def save(self, output_file_path):
        """Saves the image with the magnified insets to the file at the 
           specified path.
        """
        imageio.imsave(output_file_path, self.get_image_with_magnified_insets())
