//  Copyright (C) 2021, Christoph Peters, Karlsruhe Institute of Technology
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <https://www.gnu.org/licenses/>.


#define _CRT_SECURE_NO_WARNINGS
#include <stdio.h>
#include <stdint.h>
#include <string.h>

//! Number of vertices to be sorted
#define ENTRY_COUNT 8
//! Number of comparisons to be used in the sorting network
#define COMPARISON_COUNT 12

//! Number of networks that a single thread tries
#define NETWORKS_PER_THREAD 128
//! Number of thread blocks that are launched in a single grid
#define BLOCKS_PER_GRID 262144
//! Number of threads that are launched in a single thread block
#define THREADS_PER_BLOCK 256

//! A few macros to switch between CUDA and C easily
#define USE_CUDA 1
#if USE_CUDA
#define DEVICE_MEMORY __device__
#define DEVICE_FUNCTION __device__
#define GLOBAL_FUNCTION __global__
#else
#define DEVICE_MEMORY
#define DEVICE_FUNCTION static inline
#define GLOBAL_FUNCTION static inline
#undef BLOCKS_PER_GRID
#define BLOCKS_PER_GRID 1
#undef THREADS_PER_BLOCK
#define THREADS_PER_BLOCK 1
#endif


//! The minimal depth of a working sorting network encountered thus far
DEVICE_MEMORY uint32_t min_sorting_network_depth = 0xffffffff;


#if ENTRY_COUNT == 3
#define PERMUTATION_COUNT 6
#define CODED_IDENTITY 136ul
#define PAIR_COUNT 3
DEVICE_MEMORY const uint32_t permutations[PERMUTATION_COUNT] = {
	66ul, 80ul, 136ul, 17ul, 10ul, 129ul
};
DEVICE_MEMORY const uint8_t comparisons[PAIR_COUNT][2] = {
	{0, 1}, {0, 2}, {1, 2}
};
#elif ENTRY_COUNT == 4
#define PERMUTATION_COUNT 16
#define CODED_IDENTITY 1672ul
#define PAIR_COUNT 6
DEVICE_MEMORY const uint32_t permutations[PERMUTATION_COUNT] = {
	83ul, 209ul, 1224ul, 1035ul, 1217ul, 1546ul, 1672ul, 1091ul, 720ul, 153ul, 90ul, 1602ul, 531ul, 664ul, 1665ul, 538ul
};
DEVICE_MEMORY const uint8_t comparisons[PAIR_COUNT][2] = {
	{0, 1}, {0, 2}, {0, 3}, {1, 2}, {1, 3}, {2, 3}
};
#elif ENTRY_COUNT == 5
#define PERMUTATION_COUNT 40
#define CODED_IDENTITY 18056ul
#define PAIR_COUNT 10
DEVICE_MEMORY const uint32_t permutations[PERMUTATION_COUNT] = {
	675ul, 9985ul, 18056ul, 2257ul, 8284ul, 12820ul, 9992ul, 1809ul, 12372ul, 668ul, 13380ul, 5904ul, 17930ul, 14472ul, 17475ul, 14346ul, 6352ul, 17986ul, 17419ul, 8739ul, 14402ul, 794ul, 8291ul, 18049ul, 16467ul, 13324ul, 1249ul, 1305ul, 10440ul, 5344ul, 4252ul, 10433ul, 16915ul, 4378ul, 4322ul, 8732ul, 738ul, 4259ul, 5400ul, 14465ul
};
DEVICE_MEMORY const uint8_t comparisons[PAIR_COUNT][2] = {
	{0, 1}, {0, 2}, {0, 3}, {0, 4}, {1, 2}, {1, 3}, {1, 4}, {2, 3}, {2, 4}, {3, 4}
};
#elif ENTRY_COUNT == 6
#define PERMUTATION_COUNT 96
#define CODED_IDENTITY 181896ul
#define PAIR_COUNT 15
DEVICE_MEMORY const uint32_t permutations[PERMUTATION_COUNT] = {
	80641ul, 19153ul, 47440ul, 15121ul, 66347ul, 43800ul, 102565ul, 180755ul, 66403ul, 139357ul, 80193ul, 120961ul, 43352ul, 98981ul, 98988ul, 5356ul, 5475ul, 42792ul, 34021ul, 152643ul, 139805ul, 153224ul, 168092ul, 6490ul, 181315ul, 177164ul, 164508ul, 14673ul, 34658ul, 143445ul, 34028ul, 66284ul, 69987ul, 69861ul, 144397ul, 69931ul, 102572ul, 143893ul, 51920ul, 42848ul, 84673ul, 107045ul, 120968ul, 120842ul, 106597ul, 10081ul, 181896ul, 181770ul, 176212ul, 181826ul, 80200ul, 5986ul, 144453ul, 181259ul, 177220ul, 35162ul, 172572ul, 11033ul, 153098ul, 47888ul, 34602ul, 117258ul, 34147ul, 10025ul, 10585ul, 117377ul, 88264ul, 22737ul, 153154ul, 5349ul, 5930ul, 6938ul, 107052ul, 117384ul, 135325ul, 69868ul, 120898ul, 35610ul, 34091ul, 176660ul, 5419ul, 80648ul, 153217ul, 180307ul, 88257ul, 181889ul, 152083ul, 131741ul, 84680ul, 151635ul, 117314ul, 172124ul, 66277ul, 152587ul, 106604ul, 55504ul
};
DEVICE_MEMORY const uint8_t comparisons[PAIR_COUNT][2] = {
	{0, 1}, {0, 2}, {0, 3}, {0, 4}, {0, 5}, {1, 2}, {1, 3}, {1, 4}, {1, 5}, {2, 3}, {2, 4}, {2, 5}, {3, 4}, {3, 5}, {4, 5}
};
#elif ENTRY_COUNT == 7
#define PERMUTATION_COUNT 224
#define CODED_IDENTITY 1754760ul
#define PAIR_COUNT 21
DEVICE_MEMORY const uint32_t permutations[PERMUTATION_COUNT] = {
	1376998ul, 820526ul, 941122ul, 277858ul, 1446046ul, 1754123ul, 941066ul, 387856ul, 1525384ul, 1578213ul, 1708189ul, 44387ul, 530348ul, 1455118ul, 1114862ul, 276906ul, 1740956ul, 646408ul, 1237523ul, 820596ul, 1524243ul, 1151157ul, 272755ul, 452816ul, 714945ul, 1671845ul, 1754753ul, 852789ul, 1516060ul, 560483ul, 43379ul, 649985ul, 42924ul, 42805ul, 1754760ul, 646401ul, 355096ul, 1754634ul, 1344742ul, 1155630ul, 1749076ul, 714952ul, 81249ul, 1754179ul, 530229ul, 272811ul, 856436ul, 558901ul, 856373ul, 118097ul, 219345ul, 1238657ul, 1409702ul, 1155182ul, 84889ul, 1082613ul, 1753619ul, 43435ul, 1525314ul, 1450526ul, 1118453ul, 276850ul, 1266755ul, 679112ul, 1238664ul, 416976ul, 560035ul, 1238083ul, 1266195ul, 558964ul, 380240ul, 52122ul, 1450078ul, 47474ul, 558894ul, 856492ul, 1454166ul, 1525377ul, 289562ul, 281498ul, 642369ul, 559020ul, 944769ul, 791854ul, 1675429ul, 1523795ul, 944650ul, 976962ul, 641921ul, 1679461ul, 1606885ul, 1750028ul, 743617ul, 531363ul, 1380582ul, 1749524ul, 1744988ul, 56602ul, 682696ul, 1520148ul, 530222ul, 1524803ul, 1679909ul, 343392ul, 1237075ul, 1151150ul, 791861ul, 649992ul, 1704605ul, 1524747ul, 679105ul, 1454614ul, 1114869ul, 1737372ul, 1507996ul, 852908ul, 347032ul, 1082606ul, 277410ul, 117649ul, 682689ul, 791980ul, 1238538ul, 1442462ul, 1642725ul, 977025ul, 342440ul, 976906ul, 743624ul, 1717317ul, 641928ul, 1417766ul, 977032ul, 1717261ul, 1515612ul, 1753171ul, 52570ul, 791924ul, 43939ul, 420560ul, 944706ul, 272300ul, 190673ul, 1155637ul, 944776ul, 559531ul, 1155189ul, 1520652ul, 92953ul, 1413286ul, 342384ul, 530859ul, 1053934ul, 379792ul, 941192ul, 158417ul, 1238027ul, 852782ul, 1750084ul, 47530ul, 1745436ul, 1005578ul, 342944ul, 351512ul, 273763ul, 852852ul, 80801ul, 85337ul, 1316070ul, 1716757ul, 820652ul, 273315ul, 1716309ul, 1005704ul, 1511580ul, 48034ul, 1639141ul, 856366ul, 154833ul, 42798ul, 281946ul, 1266699ul, 820533ul, 1417318ul, 272244ul, 42868ul, 1005697ul, 481488ul, 559475ul, 80297ul, 1238594ul, 1455174ul, 1712221ul, 1005634ul, 1147566ul, 642376ul, 1267266ul, 1267336ul, 1754690ul, 1267329ul, 1712669ul, 1265747ul, 1267210ul, 530803ul, 272181ul, 530292ul, 1147573ul, 60186ul, 1520708ul, 125713ul, 80241ul, 1525258ul, 1519700ul, 941185ul, 285978ul, 122129ul, 1118446ul, 48482ul, 1053941ul, 347480ul, 384272ul, 531811ul, 272174ul, 89369ul
};
DEVICE_MEMORY const uint8_t comparisons[PAIR_COUNT][2] = {
	{0, 1}, {0, 2}, {0, 3}, {0, 4}, {0, 5}, {0, 6}, {1, 2}, {1, 3}, {1, 4}, {1, 5}, {1, 6}, {2, 3}, {2, 4}, {2, 5}, {2, 6}, {3, 4}, {3, 5}, {3, 6}, {4, 5}, {4, 6}, {5, 6}
};
#elif ENTRY_COUNT == 8
#define PERMUTATION_COUNT 512
#define CODED_IDENTITY 16434824ul
#define PAIR_COUNT 28
DEVICE_MEMORY const uint32_t permutations[PERMUTATION_COUNT] = {
	4241783ul, 352163ul, 16430148ul, 4246459ul, 2817304ul, 7568008ul, 11539710ul, 388962ul, 8918910ul, 16434243ul, 8431479ul, 946001ul, 16130142ul, 10757943ul, 10212993ul, 16060646ul, 2219938ul, 16434187ul, 6334846ul, 6019265ul, 9180605ul, 2353946ul, 9209214ul, 12306056ul, 7567882ul, 7858186ul, 16433683ul, 347627ul, 13701359ul, 2739184ul, 16434824ul, 12526108ul, 11277623ul, 4471159ul, 6565356ul, 2178036ul, 16429140ul, 16425500ul, 11015998ul, 2178988ul, 982289ul, 342517ul, 4241790ul, 9241463ul, 16397325ul, 16126110ul, 2739128ul, 15835694ul, 14599816ul, 2177980ul, 16286949ul, 417242ul, 4471229ul, 12855087ul, 941969ul, 2177525ul, 12304467ul, 16417436ul, 2182635ul, 15827630ul, 14557277ul, 15996134ul, 3922128ul, 683865ul, 14562317ul, 7535626ul, 6822844ul, 9926273ul, 647073ul, 2325274ul, 9241589ul, 11604215ul, 10212930ul, 4471796ul, 13734063ul, 6822263ul, 720153ul, 3039120ul, 1824977ul, 2194787ul, 8660855ul, 7829633ul, 16397381ul, 16322789ul, 14594580ul, 13730479ul, 2252698ul, 11568375ul, 16392285ul, 5237505ul, 11640958ul, 10442369ul, 12305475ul, 11600631ul, 8660981ul, 6850935ul, 12535362ul, 392546ul, 10212363ul, 2751840ul, 641977ul, 2177399ul, 3075856ul, 15500590ul, 454426ul, 379890ul, 2878232ul, 6248648ul, 8660925ul, 348075ul, 2740136ul, 7858305ul, 8116234ul, 10441795ul, 8660862ul, 679385ul, 12300820ul, 5135809ul, 347067ul, 11636919ul, 9954826ul, 11044670ul, 14599746ul, 4246515ul, 16430092ul, 10528567ul, 15471918ul, 12305986ul, 384930ul, 6822900ul, 6851061ul, 3140368ul, 6565804ul, 5136257ul, 16122526ul, 16093350ul, 10441739ul, 12305930ul, 4471285ul, 9241533ul, 6822333ul, 15733998ul, 16425052ul, 13992615ul, 16359973ul, 8947575ul, 9180542ul, 978705ul, 2293018ul, 2183083ul, 9245117ul, 13738543ul, 14451941ul, 8431549ul, 5732545ul, 5176577ul, 6852524ul, 8918903ul, 12535425ul, 12518044ul, 16351909ul, 9925699ul, 16434754ul, 12530756ul, 9953363ul, 14598227ul, 15762670ul, 4247019ul, 6336428ul, 16433235ul, 2215850ul, 15794926ul, 3664080ul, 3337936ul, 14599179ul, 2215402ul, 14562373ul, 12525660ul, 7829640ul, 781081ul, 7596610ul, 13959911ul, 9926210ul, 347123ul, 9954945ul, 2849560ul, 10213000ul, 15831214ul, 9954315ul, 14952238ul, 342398ul, 12296284ul, 1538257ul, 2219490ul, 4242924ul, 11306302ul, 8345666ul, 1240785ul, 10442376ul, 2191203ul, 14599690ul, 16359525ul, 14025375ul, 949585ul, 2776536ul, 6852076ul, 6851572ul, 14028959ul, 5173000ul, 12301380ul, 2214842ul, 11044663ul, 14590044ul, 14599235ul, 7532161ul, 2186723ul, 16319205ul, 2781016ul, 5135816ul, 16396821ul, 5499592ul, 458010ul, 16024806ul, 14557725ul, 11633335ul, 4255587ul, 14594132ul, 5172993ul, 2178540ul, 11641399ul, 642537ul, 6334909ul, 1566929ul, 9209277ul, 6334965ul, 5761224ul, 11277630ul, 12304915ul, 6564222ul, 13963495ul, 4151504ul, 4241909ul, 2260314ul, 5176584ul, 2748256ul, 3111696ul, 7535745ul, 384482ul, 490266ul, 14524517ul, 9209333ul, 4251555ul, 5499585ul, 12534795ul, 14423269ul, 14553245ul, 6564215ul, 8947582ul, 11338551ul, 6019272ul, 10212874ul, 14037527ul, 5143880ul, 1043217ul, 4241853ul, 12530700ul, 11342142ul, 12529748ul, 16421020ul, 7532098ul, 14033439ul, 12305419ul, 16097830ul, 9180661ul, 16089766ul, 14561365ul, 2252250ul, 11604222ul, 2289434ul, 13898983ul, 3043152ul, 379834ul, 5143873ul, 10212419ul, 4242364ul, 380842ul, 5435073ul, 4488547ul, 3402448ul, 13141807ul, 1244369ul, 7568001ul, 8947645ul, 2177406ul, 14000679ul, 6851005ul, 14549661ul, 1305297ul, 4480931ul, 6564796ul, 642985ul, 10441235ul, 5140289ul, 4247467ul, 14590492ul, 8345610ul, 15798510ul, 4484963ul, 2227554ul, 2214898ul, 518938ul, 5438657ul, 2776984ul, 16388253ul, 9245047ul, 651105ul, 16396373ul, 12533843ul, 13665519ul, 12530196ul, 8116360ul, 11015991ul, 14524965ul, 8116353ul, 2739688ul, 6564285ul, 11641406ul, 10442250ul, 13113135ul, 4475835ul, 11636926ul, 716569ul, 359779ul, 3341520ul, 16135238ul, 13697775ul, 7829514ul, 8918973ul, 6334839ul, 8345729ul, 11306295ul, 5435080ul, 10440787ul, 16258277ul, 2784600ul, 14599809ul, 9245054ul, 2054353ul, 14520485ul, 15238958ul, 4242420ul, 343532ul, 356195ul, 13996199ul, 3046736ul, 6335420ul, 9245173ul, 421722ul, 9953811ul, 2182075ul, 16392733ul, 8431605ul, 12534291ul, 7532042ul, 1014545ul, 13927655ul, 2744224ul, 6822270ul, 342461ul, 5470920ul, 642033ul, 9925643ul, 10211859ul, 7596673ul, 5208840ul, 425306ul, 9954882ul, 12301324ul, 5136264ul, 13636847ul, 4476395ul, 687449ul, 14598675ul, 7567938ul, 1276625ul, 342391ul, 7858312ul, 6335476ul, 5438664ul, 13435695ul, 6564852ul, 12521628ul, 4251107ul, 8116290ul, 12535432ul, 4243372ul, 12300372ul, 14038031ul, 14722862ul, 6822389ul, 11342135ul, 12296732ul, 343980ul, 9180535ul, 11600638ul, 16135182ul, 12534851ul, 342972ul, 6850942ul, 2182131ul, 10528574ul, 6823404ul, 4471166ul, 7535752ul, 15835246ul, 16130590ul, 646625ul, 9954371ul, 6564341ul, 14037079ul, 343028ul, 2223970ul, 10442306ul, 9209207ul, 13374767ul, 16134678ul, 6823852ul, 7596680ul, 16097382ul, 15532846ul, 9924691ul, 3038672ul, 14582428ul, 11539703ul, 5140296ul, 16384669ul, 11338558ul, 4476843ul, 752409ul, 3373776ul, 16429588ul, 16134230ul, 6248641ul, 14516901ul, 5208833ul, 7535682ul, 351715ul, 10757950ul, 15210286ul, 4259171ul, 8919029ul, 14032991ul, 4475891ul, 14000231ul, 6335980ul, 3635408ul, 15536430ul, 11640951ul, 16434817ul, 941521ul, 11568382ul, 16434698ul, 14038087ul, 13403439ul, 2743776ul, 417690ul, 9241470ul, 2813720ul, 14561813ul, 13738095ul, 5732552ul, 4472300ul, 7858242ul, 9925139ul, 2256730ul, 9926280ul, 8431486ul, 6851516ul, 4480483ul, 12625711ul, 3079440ul, 4472748ul, 11633342ul, 12288668ul, 12292252ul, 7829570ul, 9954952ul, 14586012ul, 9926154ul, 2187171ul, 16355493ul, 14484197ul, 8947701ul, 8345736ul, 14487781ul, 5237512ul, 380394ul, 2177469ul, 5470913ul, 14595140ul, 12306049ul, 10211411ul, 4471740ul, 7596554ul, 5761217ul, 14595084ul, 7532168ul, 13439279ul, 679833ul, 16057062ul, 12535306ul, 654689ul
};
DEVICE_MEMORY const uint8_t comparisons[PAIR_COUNT][2] = {
	{0, 1}, {0, 2}, {0, 3}, {0, 4}, {0, 5}, {0, 6}, {0, 7}, {1, 2}, {1, 3}, {1, 4}, {1, 5}, {1, 6}, {1, 7}, {2, 3}, {2, 4}, {2, 5}, {2, 6}, {2, 7}, {3, 4}, {3, 5}, {3, 6}, {3, 7}, {4, 5}, {4, 6}, {4, 7}, {5, 6}, {5, 7}, {6, 7}
};
#endif


//! Constructs a sorting network among all possible sorting networks with the
//! specified comparison count from a single index
DEVICE_FUNCTION void construct_sorting_network(uint8_t network[COMPARISON_COUNT][2], uint64_t index) {
	for (uint64_t i = 0; i != COMPARISON_COUNT; ++i) {
		uint64_t comparison = index % PAIR_COUNT;
		network[i][0] = comparisons[comparison][0];
		network[i][1] = comparisons[comparison][1];
		index /= PAIR_COUNT;
	}
}


//! Returns true if and only if the sorting network with the given index using
//! the given number of comparisons sorts all members of permutations correctly
DEVICE_FUNCTION bool try_sorting_network(uint8_t network[COMPARISON_COUNT][2]) {
	for (uint32_t i = 0; i != PERMUTATION_COUNT; ++i) {
		uint32_t permutation = permutations[i];
		for (uint32_t j = 0; j != COMPARISON_COUNT; ++j) {
			uint32_t lhs_shift = 3 * network[j][0];
			uint32_t rhs_shift = 3 * network[j][1];
			uint32_t lhs_mask = 7 << lhs_shift;
			uint32_t rhs_mask = 7 << rhs_shift;
			uint32_t lhs = (permutation & lhs_mask) >> lhs_shift;
			uint32_t rhs = (permutation & rhs_mask) >> rhs_shift;
			uint32_t swapped_permutation = (permutation & ~lhs_mask & ~rhs_mask) | (lhs << rhs_shift) | (rhs << lhs_shift);
			permutation = (lhs > rhs) ? swapped_permutation : permutation;
		}
		if (permutation != CODED_IDENTITY)
			return false;
	}
	return true;
}


//! Returns the depth of the given sorting network
DEVICE_FUNCTION uint32_t compute_sorting_network_depth(uint8_t network[COMPARISON_COUNT][2]) {
	uint32_t accounted = 0;
	const uint32_t all_accounted = (1 << COMPARISON_COUNT) - 1;
	uint32_t touched = 0;
	uint32_t depth = 0;
	// Find a maximal number of comparisons to execute in parallel until all
	// are accounted for
	while (accounted != all_accounted) {
		++depth;
		touched = 0;
		bool layer_begun = false;
		for (uint32_t i = 0; i != COMPARISON_COUNT; ++i) {
			uint32_t lhs_bit = (1 << network[i][0]);
			uint32_t rhs_bit = (1 << network[i][1]);
			bool collision = ((touched & lhs_bit) | (touched & rhs_bit)) != 0;
			bool add_to_layer = ((accounted & (1 << i)) == 0 && !collision);
			layer_begun |= add_to_layer;
			touched = layer_begun ? (touched | lhs_bit | rhs_bit) : touched;
			accounted = add_to_layer ? (accounted | (1 << i)) : accounted;
		}
	}
	return depth;
}


//! Tries the NETWORKS_PER_THREAD sorting networks with indices
//! offset + threadIdx.x + blockIdx.x * THREADS_PER_BLOCK + i * BLOCKS_PER_GRID * THREADS_PER_BLOCK
//! where i ranges from 0 to NETWORKS_PER_THREAD - 1. Whenever it finds one
//! that has smaller depth than any one found thus far, it prints it.
GLOBAL_FUNCTION void search_sorting_networks(uint64_t offset) {
#if USE_CUDA
	const uint64_t index_0 = offset + threadIdx.x + blockIdx.x * THREADS_PER_BLOCK;
#else
	const uint64_t index_0 = offset;
#endif
	const uint64_t stride = BLOCKS_PER_GRID * THREADS_PER_BLOCK;
	for (uint64_t i = 0; i != NETWORKS_PER_THREAD; ++i) {
		uint64_t index = index_0 + i * stride;
		uint8_t network[COMPARISON_COUNT][2];
		construct_sorting_network(network, index);
		if (try_sorting_network(network)) {
			uint32_t depth = compute_sorting_network_depth(network);
#if USE_CUDA
			uint32_t old_min_depth = atomicMin(&min_sorting_network_depth, depth);
#else
			uint32_t old_min_depth = min_sorting_network_depth;
			if (old_min_depth > depth) min_sorting_network_depth = depth;
#endif
			if (old_min_depth > depth)
				printf("Found a sorting network of depth %u: %llu\n", depth, index);
		}
	}
}


//! Prints the sorting network with the given index
DEVICE_FUNCTION void print_sorting_network(uint64_t index) {
	uint8_t network[COMPARISON_COUNT][2];
	construct_sorting_network(network, index);
	for (uint32_t j = 0; j != COMPARISON_COUNT; ++j)
		printf("(%u, %u), ", network[j][0], network[j][1]);
	printf("\nDepth: %u\n", compute_sorting_network_depth(network));
}


int main(int argc, char** argv) {
	// Uncomment the following line with USE_CUDA == 0 to decode the indices
	// printed by this program
	//print_sorting_network(3864469156127036);

	// The sorting networks tend to share common suffixes. For large networks
	// (> 11 comparisons) a complete brute force search becomes quite
	// impractical. Thus we enter educated guesses for suffixes here.
	uint8_t network_suffix[][2] = {
		 {0, 2}, {2, 3}, {0, 1}
	};
	uint64_t suffix_length = sizeof(network_suffix) / sizeof(network_suffix[0]);
	// suffix_length = 0;
	uint64_t network_count = 1;
	for (uint64_t i = 0; i != COMPARISON_COUNT - suffix_length; ++i)
		network_count *= PAIR_COUNT;
	uint64_t first_relevant_network = 0;
	uint64_t offset_factor = network_count;
	for (uint64_t i = 0; i != suffix_length; ++i) {
		uint64_t comparison_index = 0;
		for (uint8_t lhs = 0; lhs != ENTRY_COUNT; ++lhs) {
			for (uint8_t rhs = lhs + 1; rhs != ENTRY_COUNT; ++rhs) {
				if (lhs == network_suffix[i][0] && rhs == network_suffix[i][1])
					first_relevant_network += offset_factor * comparison_index;
				++comparison_index;
			}
		}
		offset_factor *= PAIR_COUNT;
	}
	// Search through all networks with this suffix
	uint64_t networks_per_grid = ((uint64_t) NETWORKS_PER_THREAD) * ((uint64_t) THREADS_PER_BLOCK) * ((uint64_t) BLOCKS_PER_GRID);
	uint64_t grid_count = (network_count + networks_per_grid - 1) / networks_per_grid;
	printf("Running %llu grids.\n", grid_count);
	for (uint64_t i = 0; i != grid_count; ++i) {
		uint64_t offset = first_relevant_network + i * networks_per_grid;
#if USE_CUDA
	    search_sorting_networks<<<BLOCKS_PER_GRID, THREADS_PER_BLOCK>>>(offset);
#else
	    search_sorting_networks(offset);
#endif
		if (grid_count > 100 && (i + 1) % (grid_count / 100) == 0)
			printf("%llu submitted.\n", i + 1);
	}
	cudaError_t error_code = cudaDeviceSynchronize();
	if (error_code != 0)
		printf("Got CUDA error code %u.\n", error_code);
	return 0;
}
