#  Copyright (C) 2021, Christoph Peters, Karlsruhe Institute of Technology
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.


import numpy as np
import itertools
from math import log, ceil, factorial
from random import shuffle
from matplotlib import pyplot


def get_legal_permutations(entry_count):
    """
    :return: A list of all permutations of range(entry_count), which may occur
        as order of vertices in a planar polygon, sorted by azimuth. They map
        polygon indices to indices in the sorted list.
    """
    legal_permutations = list()
    for permutation in itertools.permutations(range(entry_count)):
        permutation = list(permutation)
        begin = permutation.index(0)
        end = permutation.index(entry_count - 1)
        if begin < end:
            forward = permutation[begin:end]
            backward = permutation[end:] + permutation[:begin]
        else:
            forward = permutation[begin:] + permutation[:end]
            backward = permutation[end:begin]
        if all(forward[i] < forward[i + 1] for i in range(len(forward) - 1)) and \
                all(backward[i] > backward[i + 1] for i in range(len(backward) - 1)):
            legal_permutations.append(permutation)
    return legal_permutations


def get_inner_outer_flags(permutation):
    """Marks vertices as inner or outer vertices.
    :param permutation: A single legal permutation.
    :return: A list of integers 0 (inner) or 1 (outer). It is indexed by the
        values of permutation minus one and has len(permutation) - 2 entries.
    """
    result = (len(permutation) - 2) * [0]
    begin = permutation.index(0)
    end = permutation.index(len(permutation) - 1)
    if begin < end:
        backward = permutation[end + 1:] + permutation[:begin]
    else:
        backward = permutation[end + 1:begin]
    for index in backward:
        result[index - 1] = 1
    return result


def get_optimal_sorting_network(entry_count):
    """
    Returns an optimal sorting network (in terms of number of comparisons) for
    a list of the given size. The method is based on a predefined list of
    sorting networks and does not support arbitrary sizes. The bigger networks
    are from this paper, specifically pages 14 and 10:
    https://arxiv.org/abs/1405.5754
    :param entry_count: Number of elements in the list being sorted.
    :return: A list of index pairs, indicating which channels should be
        compared in each step.
    """
    if entry_count <= 1:
        return []
    elif entry_count == 2:
        return [(0, 1)]
    elif entry_count == 3:
        return [(0, 1), (0, 2), (1, 2)]
    elif entry_count == 4:
        # Source: https://hoytech.github.io/sorting-networks/
        return [(0, 2), (1, 3), (0, 1), (2, 3), (1, 2)]
    elif entry_count == 5:
        # Page 10
        return [(0, 1), (2, 3), (0, 2), (1, 3), (1, 2), (3, 4), (1, 3), (0, 1), (2, 3)]
        # Page 14
        # return [(3, 4), (2, 4), (2, 3), (0, 2), (0, 1), (1, 2), (1, 3), (2, 4), (2, 3)]
    elif entry_count == 6:
        # Page 14
        return [(0, 3), (2, 5), (0, 2), (1, 4), (0, 1), (3, 4), (2, 3), (4, 5), (1, 4), (1, 2), (3, 4), (2, 3)]
        # Page 10
        # return [(0, 1), (2, 3), (4, 5), (0, 2), (0, 4), (1, 3), (1, 5), (2, 4), (1, 2), (3, 5), (3, 4), (2, 3)]
    elif entry_count == 7:
        # Page 10
        return [(0, 1), (2, 3), (4, 5), (0, 6), (2, 4), (0, 2), (1, 3), (5, 6), (1, 4), (2, 5), (1, 2), (4, 5), (2, 4),
                (3, 6), (3, 5), (3, 4)]
        # Page 14
        # return [(1, 4), (5, 6), (0, 5), (2, 3), (1, 2), (5, 6), (1, 5), (0, 1), (3, 4), (2, 3), (1, 2), (4, 6), (2, 4), (3, 5), (2, 3), (4, 5)]
    elif entry_count == 8:
        # Page 10
        return [(0, 1), (2, 3), (4, 5), (6, 7), (0, 2), (4, 6), (1, 3), (0, 4), (5, 7), (3, 7), (1, 5), (3, 5), (2, 6),
                (2, 4), (1, 2), (3, 6), (2, 4), (5, 6), (3, 4)]


def get_specialized_sorting_network(entry_count):
    """
    Like get_optimal_sorting_network() except that the sorting networks are
    only guaranteed to work correctly on legal permutations. They have minimal
    comparison count and among such networks minimal depth up to entry_count 6.
    For entry_count 7 and 8 networks are available but they are not as optimal.
    """
    if entry_count <= 1:
        return []
    elif entry_count == 2:
        return [(0, 1)]
    elif entry_count == 3:
        return [(1, 2), (0, 2), (0, 1)]
    elif entry_count == 4:
        return [(1, 3), (0, 2), (2, 3), (0, 1)]
    elif entry_count == 5:
        return [(2, 4), (1, 3), (1, 2), (0, 3), (3, 4), (0, 2), (2, 3), (0, 1)]
    elif entry_count == 6:
        return [(3, 5), (2, 4), (1, 5), (0, 4), (1, 3), (0, 2), (0, 1), (2, 3), (4, 5)]
    elif entry_count == 7:
        return [(2, 5), (1, 6), (5, 6), (3, 4), (0, 4), (4, 6), (1, 3), (3, 5), (0, 2), (0, 1), (2, 3), (4, 5)]
    elif entry_count == 8:
        return get_bitonic_merge_network(0, entry_count)


def apply_sorting_network(sequence, network):
    """
    Sorts the given list in place using the given sorting network.
    :return List of swapping decisions (0 for no swap, 1 for a swap).
    """
    result = list()
    for lhs, rhs in network:
        if sequence[lhs] > sequence[rhs]:
            sequence[lhs], sequence[rhs] = sequence[rhs], sequence[lhs]
            result.append(1)
        else:
            result.append(0)
    return result


def test_sorting_network(entry_count, network, permutations=None):
    """
    Tests the given sorting network by applying it to every single given
    permutation.
    :param entry_count: Number elements to be sorted.
    :param network: A sorting network such as the one returned by
        get_optimal_sorting_network(entry_count).
    :param permutations: A list of permutations to try. Pass None to use all
        possible permutations.
    :return: True iff the sorting network works.
    """
    sorted_range = tuple(range(entry_count))
    if permutations is None:
        permutations = itertools.permutations(range(entry_count))
    for permutation in permutations:
        original_permutation = permutation
        permutation = list(permutation)
        apply_sorting_network(permutation, network)
        if sorted_range != tuple(permutation):
            print(original_permutation)
            print(permutation)
            return False
    return True


def search_sorting_network(entry_count, comparison_count, permutations):
    """
    This function performs a more or less brute force search for sorting
    networks that are capable of sorting the given permutations.
    :param entry_count: Number of elements to be sorted.
    :param comparison_count: Number of comparisons which should be performed by
        the sorting network.
    :param permutations: Permutations given as sequences of indices from 0 to
        entry_count - 1.
    :return: If a suitable sorting network is found, it is returned, otherwise
        the method returns None.
    """
    print("Searching a sorting network with %d comparisons for %d elements." % (comparison_count, entry_count))
    # Enumerate possible comparisons
    comparisons = list()
    for j in range(entry_count):
        for k in range(j + 1, entry_count):
            comparisons.append((j, k))
    # Construct sorting networks from them
    networks = itertools.product(*comparison_count * [comparisons])
    if entry_count >= 256:
        for network in networks:
            if test_sorting_network(entry_count, network, permutations):
                return network
    else:
        # Shuffle the permutations randomly such that a prefix is somewhat
        # representative
        permutations = list(permutations)
        shuffle(permutations)
        # Put them into a bunch of arrays to dismiss nonsense early
        splits = [int(ratio * len(permutations)) for ratio in [0.05, 0.3]]
        splits = [(0, splits[0]), (splits[0], splits[1]), (splits[1], len(permutations))]
        splits = [(lhs, rhs) for lhs, rhs in splits if lhs < rhs]
        permutation_arrays = [np.transpose(np.asarray(permutations[begin: end], dtype=np.uint8)) for begin, end in
                              splits]
        sorted_range = np.arange(entry_count)[:, np.newaxis]
        for network in networks:
            works = True
            for permutations in permutation_arrays:
                sorted_permutations = np.copy(permutations)
                for lhs, rhs in network:
                    pair = sorted_permutations[lhs:(rhs + 1):(rhs - lhs), :]
                    pair[...] = np.where((pair[0, :] > pair[1, :])[np.newaxis, :].repeat(2, 0), pair[::-1, :], pair)
                if np.any(sorted_permutations != sorted_range):
                    works = False
                    break
            if works:
                return network
    return None


def find_marking_strategy(entry_count, network):
    """
    Figures out how to mark vertices as inner or outer vertices using nothing
    but a bunch of XOR operations based on decisions in the given sorting
    network.
    :param entry_count: Number of vertices.
    :return: A 2D array of booleans of shape (entry_count - 2, network_size).
        Entry [j, k] is set if swapping decision k in the network should lead
        to a flip on the bit for vertex j - 1. These bits are initialized to
        inner.
    """
    legal_permutations = get_legal_permutations(entry_count)
    flip = np.asarray([apply_sorting_network(list(permutation), network) for permutation in legal_permutations]) == 1
    outer = np.asarray([get_inner_outer_flags(permutation) for permutation in legal_permutations]) == 1
    flip_packed = (np.where(flip, 1, 0) * (2 ** np.arange(len(network))[np.newaxis, :])).sum(1)
    outer_packed = (np.where(outer, 1, 0) * (2 ** np.arange(entry_count - 2)[np.newaxis, :])).sum(1)
    flip_packed += (flip_packed * 3) * 2 ** len(network)
    for i in range(entry_count - 2)[:0]:
        for mask in range(2 ** (len(network) * 2)):
            v = np.bitwise_and(mask, flip_packed)
            v ^= v >> 16
            v ^= v >> 8
            v ^= v >> 4
            v &= 0xf
            result = ((0x6996 >> v) & 1) != 0
            if np.all(outer[:, i] == result):
                print("Mask for %d: %s" % (i, str(mask)))
    print("uint32_t marking[PERMUTATION_COUNT][2] = {")
    shuffle = np.arange(len(legal_permutations))
    np.random.shuffle(shuffle)
    print("\t" + ", ".join(["{%d, %d}" % (flip, outer) for flip, outer in zip(flip_packed[shuffle], outer_packed[shuffle])]))
    print("};")


def print_c_tables(entry_count):
    """
    Prints source code with defines and hard-coded arrays that provide
    information about optimization problems for the C implementation.
    :param entry_count: The number of vertices for which the information should
        be printed.
    """
    legal_permutations = list(get_legal_permutations(n))
    shuffle(legal_permutations)
    print("#define PERMUTATION_COUNT %d" % len(legal_permutations))
    print("#define CODED_IDENTITY %uul" % sum([i << (3 * i) for i in range(entry_count)]))
    print("#define PAIR_COUNT %d" % ((n * (n - 1)) // 2))
    print("DEVICE_MEMORY const uint32_t permutations[PERMUTATION_COUNT] = {")
    coded_permutations = [sum([entry << (3 * i) for i, entry in enumerate(permutation)]) for permutation in legal_permutations]
    print("\t" + ", ".join(["%uul" % code for code in coded_permutations]))
    print("};")
    print("DEVICE_MEMORY const uint8_t comparisons[PAIR_COUNT][2] = {")
    print("\t" + ", ".join(["{%d, %d}" % (lhs, rhs) for lhs in range(n) for rhs in range(lhs + 1, n)]))
    print("};")


def plot_sorting_network(network, entry_count):
    """
    Plots the given sorting network using matplotlib.
    :param network: A network as returned by get_optimal_sorting_network().
    :param entry_count: Number of channels.
    """
    pyplot.xticks(np.arange(len(network)))
    pyplot.yticks(np.arange(entry_count))
    pyplot.grid(axis="y")
    for j, comparison in enumerate(network):
        pyplot.plot([j, j], [comparison[0], comparison[1]], "-o", color=3 * [0.5])


def get_bitonic_merge_network(begin, count, increasing=True):
    """Returns a bitonic merge network operating on channels from begin to
       begin + count - 1."""
    result = []
    if count > 1:
        power_2_count = 1
        while power_2_count < count:
            power_2_count *= 2
        power_2_count //= 2
        for i in range(begin, begin + count - power_2_count):
            result += [(i, i + power_2_count) if increasing else (i + power_2_count, i)]
        result += get_bitonic_merge_network(begin, power_2_count, increasing)
        result += get_bitonic_merge_network(begin + power_2_count, count - power_2_count, increasing)
    return result


def get_bitonic_sort_network(begin, count, increasing=True):
    """Returns a bitonic merge sort network operating on channels from begin to
       begin + count - 1."""
    result = []
    if count > 1:
        half_count = count // 2
        result += get_bitonic_sort_network(begin, half_count, not increasing)
        result += get_bitonic_sort_network(begin + half_count, count - half_count, increasing)
        result += get_bitonic_merge_network(begin, count, increasing)
    return result


if __name__ == "__main__":
    for lhs, rhs in get_specialized_sorting_network(8):
        print("compare_and_swap(polygon, %d, %d);" % (lhs, rhs))
    # n = 7
    # legal_permutations = get_legal_permutations(n)
    """
    for plot_n in range(3, 9):
        pyplot.subplot(2, 3, plot_n - 2)
        print(plot_n, test_sorting_network(plot_n, get_specialized_sorting_network(plot_n), get_legal_permutations(plot_n)))
        plot_sorting_network(get_specialized_sorting_network(plot_n), plot_n)
    """
    """
    network = search_sorting_network(n, 7, legal_permutations)
    print(network)
    """
    """
    for n in range(3, 9):
        if n == 3:
            print("#if ENTRY_COUNT == %d" % n)
        else:
            print("#elif ENTRY_COUNT == %d" % n)
        print_c_tables(n)
    print("#endif")
    """
    """
    for permutation in legal_permutations:
        print(get_inner_outer_flags(permutation), apply_sorting_network(list(permutation), get_specialized_sorting_network(n)))
    bits = np.asarray([apply_sorting_network(list(permutation), get_specialized_sorting_network(n)) for permutation in legal_permutations])
    print(np.all(bits == 0, axis=0))
    print(np.all(bits == 1, axis=0))
    """
    """
    print(factorial(n))
    # print(log(factorial(n)) / log(2.0))
    print(ceil(log(factorial(n)) / log(2.0)))
    print(len(legal_permutations))
    # print(log(len(legal_permutations)) / log(2.0))
    print(ceil(log(len(legal_permutations)) / log(2.0)))
    """
    pyplot.show()
