#  Copyright (C) 2021, Christoph Peters, Karlsruhe Institute of Technology
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.


import re


def translate_glsl_to_cpp(source_file, destination_file, global_declarations="", function_flags="inline", use_cuda=False):
    """
    Applies some primitive substitutions that are just capable enough to turn
    GLSL code for my sampling methods into C++ code using GLM. All function
    definitions are prefixed with the given flags. If use_cuda is True, a few
    functions are renamed to match CUDA intrinsic names.
    """
    # Read input
    code = open(source_file, "r").read()
    # Includes and globals
    code = global_declarations + "\n" + code
    code = "#include \"glsl_math.h\"\n" + code
    # Flag all functions
    code = re.sub(r"^(\w+\s+\w+\s*\(.*\)\s*{)", r"%s \1" % function_flags, code, flags=re.MULTILINE)
    # Turn output parameters into non-const references
    code = re.sub(r"(\W)out\s+(\w+)\s+(\w+)(,|\))", r"\1\2& \3\4", code)
    code = re.sub(r"(\W)inout\s+(\w+)\s+(\w+)(,|\))", r"\1\2& \3\4", code)
    code = re.sub(r"(\W)inout\s+(\w+)\s+(\w+)(\s*\[.*?\])(,|\))", r"\1\2 \3\4\5", code)
    # Inline declarations of vectors work differently
    code = re.sub(r"(\W)(\w?vec[0-9])\(", r"\1make_\2(", code)
    # Replace [[unroll]] directives by #pragma unroll and erase [[dont_unroll]]
    code = re.sub(r"^([\t ]*)\[\[unroll\]\]", r"\1#pragma unroll", code, flags=re.MULTILINE)
    code = re.sub(r"^([\t ]*)\[\[dont_unroll\]\]", r"", code, flags=re.MULTILINE)
    # Rename functions
    glsl_cuda_function_names = [
        ("abs", "fabsf"),
        ("inversesqrt", "rsqrtf"),
        ("sqrt", "sqrtf"),
        ("acos", "acosf"),
        ("cos", "cosf"),
        ("sin", "sinf"),
        ("fma", "fmaf"),
        ("max", "fmaxf"),
        ("min", "fminf"),
    ]
    if use_cuda:
        for glsl_name, cuda_name in glsl_cuda_function_names:
            code = re.sub(r"(\W)%s(\s*\()" % glsl_name, r"\1%s\2" % cuda_name, code, flags=re.MULTILINE)
    # Write output
    open(destination_file, "w").write(code)


def translate_glsl_to_python(source_file, destination_file, global_declarations=""):
    """
    Applies some primitive substitutions that are just capable enough to turn
    GLSL code for my sampling methods into Python code. Global declarations are
    pasted into the output file at the beginning.
    """
    # Read input
    code = open(source_file, "r").read()
    # Changed identifiers of a few constants
    code = code.replace("M_PI_2", "2.0 * pi")
    code = code.replace("M_PI", "pi")
    code = code.replace("M_HALF_PI", "0.5 * pi")
    code = code.replace("M_INFINITY", "inf")
    code = code.replace("true", "True")
    code = code.replace("false", "False")
    # Includes and globals
    code = "from glsl_math import *\n" + global_declarations + "\n" + code
    # Remove all comments
    code = re.sub(r"[ \t]*//.[^/].*\n", "", code)
    code = re.sub(r"[ \t]*/\*.*?\*/\n", "", code, flags=re.DOTALL)
    # Remove [branch] directives and the like
    code = re.sub(r"[\t ]*\[\[flatten\]\]\n", "", code)
    code = re.sub(r"[\t ]*\[\[dont_flatten\]\]\n", "", code)
    code = re.sub(r"[\t ]*\[\[unroll\]\]\n", "", code)
    code = re.sub(r"[\t ]*\[\[dont_unroll\]\]\n", "", code)

    # Turn structs into objects that initialize everything with zeros
    def port_struct(struct_match):
        name = struct_match.group(1)
        member_list = [(member_match.group(3), member_match.group(1), member_match.group(7), member_match.group(9)) for
                       member_match in
                       re.finditer(r"((vec([0-9]?))|(uint)|(float)|(\w+))\s+(\w+)(\[([^\]]+)\])?;",
                                   struct_match.group(2))]
        declaration_list = list()
        for vector, type_match, member, array_size in member_list:
            declaration = "\t\tself.%s = " % member
            if vector is not None and len(vector) > 0:
                rhs = "zeros((%s,))" % vector
            elif type_match == "float":
                rhs = "0.0"
            elif type_match == "uint":
                rhs = "0"
            else:
                rhs = "%s()" % type_match
            if array_size is not None:
                rhs = "[%s for _i in range(%s)]" % (rhs, array_size)
            declaration += rhs
            declaration_list.append(declaration)
        return "class %s:\n\tdef __init__(self):\n" % name + "\n".join(declaration_list)

    code = re.sub(r"struct (\w+)\s*{(.*?)};", port_struct, code, flags=re.DOTALL)
    # Turn switch statements into if-else cascades
    def port_switch(switch_match):
        switch_variable = switch_match.group(1)
        switch_body = switch_match.group(2)
        switch_body = re.sub("case\s+([0-9]+):", r"if %s == \1:" % switch_variable, switch_body)
        switch_body = re.sub("\s+break;", r"", switch_body)
        switch_body = re.sub("\s+default:", r"", switch_body)
        return switch_body

    code = re.sub(r"switch\s*\(\s*(.+?)\s*\)\s*{(.*?)}", port_switch, code, flags=re.DOTALL)
    # Remove the f on float literals and the u on uint literals
    code = re.sub(r"\.([0-9]+)f", r".\1", code)
    code = re.sub(r"\.([0-9]+e[+-]?[0-9]+)f", r".\1", code)
    code = re.sub(r"(\W[0-9]+)u", r"\1", code)
    code = re.sub(r"(\W0x[0-9A-Fa-f]+)u", r"\1", code)
    # Swap indices on matrices and 2D arrays
    code = re.sub(r"(\w+)\[([^\]]+)\]\[([^\]]+)\]", r"\1[\3][\2]", code)
    # Use indices instead of x, y, z, w, etc.
    for swizzle_string in ["xyzw", "rgba", "stpq"]:
        for i, coordinate in enumerate(swizzle_string):
            code = re.sub(r"\.%s(\W)" % coordinate, r"[%d]\1" % i, code)
        for j in range(3):
            for k in range(j + 2, 4):
                code = re.sub(r"\.%s(\W)" % swizzle_string[j:k], r"[%d:%d]\1" % (j, k), code)
    # Remove type specifications on initializations
    code = re.sub(r"\t\w+ (\w+)\s*=", r"\t\1 =", code)
    # Remove basic type declarations
    code = re.sub(r"\t+(float|bool|uint) (\w+);\n", r"", code)
    # Change float vector (array) declarations
    code = re.sub(r"\tvec([0-9]) (\w+);", r"\t\2 = zeros((\1,))", code)
    code = re.sub(r"\tvec([0-9]) (\w+)\[(.*?)\]\[(.*?)\];",
                  r"\t\2 = [[zeros((\1,)) for _j in range(\4)] for _i in range(\3)]", code)
    code = re.sub(r"\tvec([0-9]) (\w+)\[(.*?)\];", r"\t\2 = [zeros((\1,)) for _i in range(\3)]", code)
    # Change float array declarations
    code = re.sub(r"\tfloat (\w+)\[(.*?)\]\[(.*?)\];", r"\t\1 = [[0.0 for _j in range(\3)] for _i in range(\2)]", code)
    code = re.sub(r"\tfloat (\w+)\[(.*?)\];", r"\t\1 = [0.0 for _i in range(\2)]", code)
    # Change struct declarations
    code = re.sub(r"\t(\w+) (\w+);", r"\t\2 = \1()", code)
    code = re.sub("\t(\w+)\s*=\s*return\(\)", r"\treturn \1", code)
    # Turn inline array declarations into lists
    code = re.sub(r"\w+\s+(\w+)\s*\[[^\]]*\]\s*=\s*{([^\}]+)};", r"\1 = [\2]", code, flags=re.DOTALL)
    def port_2d_array_declaration(array_match):
        identifier = array_match.group(1)
        array_content = array_match.group(2).replace("{", "[").replace("}", "]")
        return r"%s = asarray([%s]).T" % (identifier, array_content)

    code = re.sub(r"\w+\s+(\w+)\s*\[[^\]]*\]\[[^\]]*\]\s*=\s*{(.*?)};", port_2d_array_declaration, code, flags=re.DOTALL)
    # Replace in place array declarations by list assignments
    code = re.sub(r"\t\w+\s+(\w+)\[[0-9]+\]\s*=\s*{(.*)};", r"\t\1 = [\2]", code)
    # Replace braces by colons
    for i in range(5):
        code = re.sub(r"\s*{([^{]*?)\n[ \t]*}", r":\1", code, flags=re.DOTALL)
    # Sometimes a line ends with a break statement
    code = re.sub(r"(([\t ]+)if\s*\(.*?)\s+break;$", r"\1:\n\2\tbreak", code, flags=re.MULTILINE)
    # Else if is elif
    code = re.sub("else if", "elif", code)
    # Translate basic for loops
    code = re.sub(r"for\s*\(u?int (\w+)\s*=\s*(\w+)\s*;\s*\w+\s*!=\s*([^;]+?)\s*;\s*\+\+\w+\s*\)",
                  r"for \1 in range(\2, \3)", code)

    # Remove typenames from function declarations
    def fix_function(function_match):
        parameter_list = re.sub(r"((in)?out )?\w+ (\w+)(\[[^,]+\])?", r"\3", function_match.group(2))
        return "def %s(%s):" % (function_match.group(1), parameter_list)

    code = re.sub(r"^\w+ (\w+)\((.*)\):", fix_function, code, flags=re.MULTILINE)
    # Change the syntax of conditional operators
    code = re.sub(r"(([^><]=)|return )\s*([^\n]+)\?\s*([^;]+):\s*([^;]+)\s*;", r"\1 where(\3,\4,\5)", code,
                  flags=re.DOTALL)
    # Adapt logical operations
    code = re.sub(r"\s*&&\s*", r" and ", code)
    code = re.sub(r"\s*\|\|\s*", r" or ", code)
    code = re.sub(r"\s*!(\w)", r" not \1", code)
    # Rename a few functions
    code = re.sub(r"(\W)min\(", r"\1minimum(", code)
    code = re.sub(r"(\W)max\(", r"\1maximum(", code)
    code = re.sub(r"(\W)asin\(", r"\1arcsin(", code)
    code = re.sub(r"(\W)acos\(", r"\1arccos(", code)
    code = re.sub(r"(\W)atan\((.*),", r"\1arctan2(\2,", code)
    code = re.sub(r"(\W)atan\(", r"\1arctan(", code)
    code = re.sub(r"(\W)clamp\(", r"\1clip(", code)
    code = re.sub(r"(\W)outerProduct\(", r"\1outer(", code)
    code = re.sub(r"(\W)determinant\(", r"\1linalg.det(", code)
    # A few special substitutions
    code = re.sub(r"vec3\s*\(\s*([0-9]+).0,\s*([0-9]+).0\s*,\s*([0-9]+).0\s*\)", r"asarray([\1.0,\2.0,\3.0])", code)
    code = re.sub(r"(cholesky\)?\s*)\*", r"\1@", code)
    # Use spaces
    code = re.sub(r"\t", "    ", code)
    # Never use semicolons
    code = re.sub(r";", "", code)
    # Remove includes
    code = re.sub(r"#include.*", "", code)
    # Write output
    open(destination_file, "w").write(code)


if __name__ == "__main__":
    globals = dict(MIN_POLYGONAL_LIGHT_VERTEX_COUNT=3, MAX_POLYGONAL_LIGHT_VERTEX_COUNT=4, MAX_POLYGON_VERTEX_COUNT=5)
    globals_python = "\n".join(["%s = %d" % item for item in globals.items()])
    translate_glsl_to_python("../vulkan_renderer/src/shaders/polygon_sampling.glsl",
                             "polygon_sampling_ported.py", globals_python)
    globals_cpp = "\n".join(["#define %s %d" % item for item in globals.items()])
    translate_glsl_to_cpp("../vulkan_renderer/src/shaders/polygon_sampling.glsl",
                          "find_worst_case/polygon_sampling_ported.h", globals_cpp, "inline", False)
    translate_glsl_to_python("../vulkan_renderer/src/shaders/cubic_solver.glsl", "cubic_solver_ported.py")
    globals_python = "from cubic_solver_ported import solve_cubic\n\n" + globals_python
    globals_python = "from polygon_sampling_ported import *\n" + globals_python
    translate_glsl_to_python("../vulkan_renderer/src/shaders/polygon_sampling_related_work.glsl",
                             "polygon_sampling_related_work_ported.py", globals_python)
    # To find the worst case error, we may want to use double precision
    with open("find_worst_case/polygon_sampling_ported.h", "r") as file:
        cpp_code = file.read()
    cpp_code = re.sub(r"(\W)float(\W)", r"\1Float\2", cpp_code)
    with open("find_worst_case/polygon_sampling_ported.h", "w") as file:
        file.write(cpp_code)
    print("Done.")
